/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code.  The copyright
notice included with the software must be maintained in all copies
produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2008, John Conover, All Rights Reserved.

The CM17A section of the code is based on code from the BottleRocket
program, (http://www.linuxha.com/bottlerocket/,) by Tymm Twillman,
Ashley Clark, David Anderson, Warner Losh, and, Christian Gafton. (c)
1999 Ashley Clark (aclark@ghoti.org) and Tymm Twillman (tymm@acm.org)
Free Software.  LGPL applies.  No warranties expressed or implied.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

    http://www.johncon.com/

------------------------------------------------------------------------------

spx10S0.c is for asynchronous raw serial port communication with the
X10 CM17A Computer Interface and MR26A RF Receiver used in many home
automation systems.  The program is intended to provide simple and
reliable serial port communication between a computer running Linux
and the X10 CM17A, (e.g., "Firecracker,") Computer Interface and MR26A
RF Receiver, (both on the same port, /dev/ttyS0,) and allows X10
devices to be controlled, and monitored, via RF transmission by the
computer, (the X10 UR81A Universal Remote, often distributed with the
MR26A is supported, also.)

The default port is /dev/ttyS0 at 9600 baud. The default semaphore
lock file is /var/lock/LCK..ttyS0, which has an fcntl(2) advisory lock
applied. The serial port is controlled through termios(3), and is
restored to original values on exit. The serial port child read
process is created by fork(2) from the parent write process, and both
processes respond to SIGINT = ^C and SIGTERM signals. Asynchronous
serial port control is implemented with select(2). The input file,
(usually stdin for commands,) uses canonical input for flow
control. The X10 device and transmission codes are cross indexed in a
binary search database using mmap(2), which also contains a command
sequence field, (including shell calls,) that are executed in response
to a specific transmission code. Variable names and states are
supported via a hash lookup table.

The database file is a Unix tab delimited field file, in lexical
order, (i.e., constructed with sort(1),) consisting of four fields:

    1) The X10 transmission code.

    2) The device code and command.

    3) A user defined alias.

    4) A command sequence to execute in response to the transmission
       code.

For example:

    d5 aa 64 00 ad  a9 on   porch motion  a11 on; \
        isset ix exec echo "motion detected" \
        | mailx -s "motion detected" 5551234@myphone.com

would turn on device A11, and then, send a message to my cell phone if
motion was detected on my front porch, but only if the variable "ix"
was set, (e.g., the variable "ix" enables transmission to my cell
phone.)  The motion detector is on house code A, device code 9, which
has the transmission sequence, "d5 aa 64 00 ad." (The reason for
including the transmission sequence in the database was for
extensibility with other devices and products-the program will print
unknown transmission sequences received for future inclusion in the
database-or take evasive action in jamming or collision situations,
etc.)

The language in field 4:

    [a-p][1-9][0-6]?: house code and unit code, i.e., "b12"

    on: turn a device on, i.e., "b12 on"

    off: turn a device off, i.e., "b12 off"

    dim ([0-9][0-2]?)?: dim a device, i.e., "b12 dim" or "b12 dim 5"

    bright  ([0-9][0-2]?)?: brighten a device, i.e., "b12 bright" or "b12 bright 5"

    all_off: all devices off in a house, i.e., "b all_off"

    all_on: all devices on in a house, i.e., "b all_on"

    all_lamps_off: all lamps off in a house, i.e., "b all_lamps_off", (actually an alias for all_off)

    all_lamps_on: all lamps on in a house, i.e., "b all_lamps_on", (actually an alias for all_on)

    quit: stop the program

    exit: alias for "quit"

    set <variable>: set a variable, i.e., "set x123", (a variable is either set, or it isn't)

    reset <variable>: reset a variable, i.e., "reset x123", (a variable is either set, or it isn't)

    unset, alias for "reset"

    isset <variable> <action>: test if a variable is set, i.e., "isset x123 b12 on", (a variable is either set, or it isn't)

    isnotset <variable> <action>: test if a variable is not set, i.e., "isnotset x123 b12 on", (a variable is either set, or it isn't)

    timeset <variable>: the last time a variable was set, in ctime(3)

    istimeset <variable> seconds action: if a variable has been set at least seconds many seconds, (or has been set less than seconds many seconds if seconds is negative,) do action

    exec <command>: execute a command

    system, alias for "exec"

    variable, any alphanumeric string not one of the above keywords

    ';': command delimiter, i.e., "isset x123 b12 on; isset abc b11 off"

Note that the isset and isnotset commands allow simple constructs-the
main function is to inhibit spawning a shell for every instance of a
state change, (like a motion detector in a room,) for example:

    isset x123 exec myfile; isset abc exec myprogram

would invoke a shell process only if x123, (or abc,) is set; any more
complicated constructs should be handled by the shell interpreter in
a batch file.

The isset and isnotset commands can be chained to provide "and"
boolean functionality, for example:

    isset x123 isnotset y456 exec myfile

would invoke a shell process only if x123 is set, and y456 is not set.

The program is not intended to be interactive, (although it does
accept commands from stdin,) and is a network centric automaton,
(daemon, 'bot., etc.,) launched in the computer's init scripts.  An
example invocation is:

    mknod myfile p
    spx10S0 -f spx10S0db -i myfile

Commands, to be sent to X10 devices are written to the named pipe,
myfile, (the named pipe is also used by the program to send commands
in response to transmission sequences-fcntl(2) advisory locking is
provided for concurrency control,) and received transmissions, and
database lookups, are printed to stdout for logging considerations.

The program is intended for inclusion in a network centric embedded
home automation system, (based on a "live CD" version of Knoppix,
with no hard disk for security considerations.)

The CM17A section of the code is based on code from the BottleRocket
program, (http://www.linuxha.com/bottlerocket/,) by Tymm Twillman,
Ashley Clark, David Anderson, Warner Losh, and, Christian Gafton. (c)
1999 Ashley Clark (aclark@ghoti.org) and Tymm Twillman (tymm@acm.org)
Free Software.  LGPL applies.  No warranties expressed or implied.

The serial port code is based on the spcid program,
(http://www.johncon.com/john/cid/,) the database code is based on the
NdeX program, (http://www.johncon.com/ndex/,) and the hash table code
on hash.shar.gz, (http://www.johncon.com/.)

$Revision: 0.0$
$Date: 2008/01/07 21:31:06 $
$Id: spx10S0.c,v 0.0 2008/01/07 21:31:06 john Exp $
$Log: spx10S0.c,v $
Revision 0.0  2008/01/07 21:31:06  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <malloc.h>
#include <termios.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/mman.h>

#define _POSIX_SOURCE 1 /* POSIX compliant source */

#define BAUDRATE B9600 /* baud rate for serial port communication */
#define SPDEVICE "/dev/ttyS0" /* default serial port device */
#define LOCKFILE "/var/lock/LCK..ttyS0" /* default serial port lock file name */
#define DBFILE "spx10S0db" /* default database file name */

#define BUFLEN BUFSIZ /* size of record buffers */

#define TOKEN_SEPARATORS " \t\n\r\b," /* file record field separators */

#define VALIDATION 3 /* default int validation; number of sequential identical records for a validated transmission from the MR26 */

#define INTERBITDELAY 1500 /* default bit time for the CM17A, 500 usec., as per the X10 specification recommended minimum value */
#define PRECMDDELAY 700000 /* default CM17A time to hold output lines in "clock" position before a command */
#define POSTCMDDELAY 1200000 /* default CM17A time to stay in "clock" postion after a command */

static char rcsid[] = "$Id: spx10S0.c,v 1.12 2008/01/07 21:31:06 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2008, John Conover, All Rights Reserved"; /* the copyright banner */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Serial port ttyS0 X10 communication\n",
    "Usage: spx10S0 [-d device] [-f dbfilename] [-i infilename] [-l lockfile]\n",
    "               [-o outfilename] [-q queryfilename] [-V] [-v]\n",
    "    -d device, device name, (/dev/ttyS0)\n",
    "    -f dbfilename, database file name, (spx10S0db)\n",
    "    -i infilename, input filename, (stdin)\n",
    "    -l lockfile, lock file name, (/var/lock/LCK..ttyS0)\n",
    "    -o outfilename, output file name, (stdout)\n",
    "    -q queryfilename, query file name, (stderr)\n",
    "    -V, verbose diagnostics\n"
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening serial device\n",
    "Error getting serial device attributes\n",
    "Error installing signal handler\n",
    "Error flushing serial device attributes\n",
    "Error setting serial device attributes\n",
    "Error writing to serial device\n",
    "Error resetting serial device attributes\n",
    "Error closing serial device\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error creating lock file\n",
    "Error removing lock file\n",
    "Error forking to child read process\n",
    "Error reading from serial device\n",
    "Error locking lock file\n",
    "Error writing to lock file\n",
    "Error opening output file\n",
    "Error closing output file\n",
    "Error reading input file\n",
    "Error writing to output file\n",
    "Error enabling blocking on serial device\n",
    "Error flushing output file\n",
    "Error setting baud rate\n",
    "Error flushing serial device\n",
    "Error write buffer overflow\n",
    "Error select\n",
    "Error opening database file\n",
    "Error closing database file\n",
    "Error in status of database file\n",
    "Error mapping database file\n",
    "Error unmapping database file\n",
    "Error in database buffer\n",
    "Error in database file\n",
    "Error getting time\n",
    "Error select timer\n",
    "Error time of day\n",
    "Error iotcl\n",
    "Error command\n",
    "Error locking input file\n",
    "Error unlocking input file\n",
    "Error writing to input file\n",
    "Error allocating memory\n",
    "Error hash table already initilaized\n",
    "Error hash table mkhash () failure\n",
    "Error hash table key not found\n",
    "Error x10 command\n",
    "Error system call\n",
    "Error waiting for system call\n",
    "Error opening query file\n",
    "Error closing query file\n",
    "Error writing to query file\n"
};

#define NOERROR 0 /* negative error values, one for each index in the error message array */
#define EARGS -1
#define EOPEN -2
#define EGATTR -3
#define ESIG -4
#define EFLUSH -5
#define ESATTR -6
#define EWRITE -7
#define ERATTR -8
#define ECLOSE -9
#define EIOPEN -10
#define EICLOSE -11
#define ELOCKF -12
#define ELOCKR -13
#define EFORK -14
#define EREAD -15
#define EFCNL -16
#define EWLCK -17
#define EOOPEN -18
#define EOCLOSE -19
#define ERDIN -20
#define EOWRITE -21
#define EBLOCK -22
#define EOFLUSH -23
#define EBAUD -24
#define EDFLUSH -25
#define EWROVFL -26
#define ESEL -27
#define EDBOPEN -28
#define EDBCLOSE -29
#define ESTAT -30
#define EMAP -31
#define EUNMAP -32
#define EBOF -33
#define EDB -34
#define ETIME -35
#define ETSEL -36
#define ETOD -37
#define EIOTCL -38
#define ECMD -39
#define ELLCK -40
#define ELUCK -41
#define EIWRITE -42
#define EALLOC -43
#define EHINIT -44
#define EHMK -45
#define EHKEY -46
#define EX10 -47
#define ESYS -48
#define EWAIT -49
#define EQOPEN -50
#define EQCLOSE -51
#define EQWRITE -52

static char serial_lockname[] = {LOCKFILE}, /* default lock file name */
            *lockname = serial_lockname, /* default lock file name */
            db[] = {DBFILE}, /* default database file name */
            serial_device[] = {SPDEVICE}, /* default serial port device name */
            EOT = (char) 0x04; /* EOT = ^D, end of transmission/transaction marker character */

static int retval = NOERROR, /* return value, assume no error */
           fd = -1, /* serial port device file descriptor */
           infile = STDIN_FILENO, /* input file descriptor */
           outfile = STDOUT_FILENO, /* output file descriptor */
           queryfile = STDERR_FILENO, /* query file descriptor */
           dbfile = -1, /* database file descriptor */
           semaphorelock = -1, /* lock file file descriptor */
           state = 0, /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
           remove_lockfile = 0, /* remove lock file, 0 = no, 1 = yes */
           validation = VALIDATION, /* number of sequential identical records for a validated transmission */
           verbose = 0, /* verbose diagnostics flag, 0 = no, 1 = yes */
           InterBitDelay = INTERBITDELAY, /* CM17A time each bit is clocked out the serial port */
           PreCmdDelay = PRECMDDELAY, /* CM17A time to hold output lines in "clock" position before a command */
           PostCmdDelay = POSTCMDDELAY; /* CM17A time to stay in "clock" postion after a command */

static pid_t pid_write = (pid_t) 0, /* pid of parent write process */
             pid_read = (pid_t) 0; /* pid of child read process */

static struct termios oldtio; /* original device attributes, restored on exit */

static struct flock lockfile; /* lock file structure */

caddr_t mapbegin = (caddr_t) 0; /* reference to beginning of mmap'ed database file */

size_t filesize; /* size of database file */

enum command /* commands, x10 commands first, local commands last, concluded with variable */
{
    ON,
    OFF,
    DIM,
    BRIGHT,
    ALL_OFF,
    ALL_ON,
    ALL_LAMPS_OFF,
    ALL_LAMPS_ON,
    QUIT,
    EXIT,
    SET,
    RESET,
    UNSET,
    ISSET,
    ISNOTSET,
    TIMESET,
    ISTIMESET,
    EXEC,
    SYSTEM,
    VARIABLE
};

static const char *keyword[] = /* command interpreter keywords, must be the same number, and in same sequence as enum command */
{
    "on",
    "off",
    "dim",
    "bright",
    "all_off",
    "all_on",
    "all_lamps_off",
    "all_lamps_on",
    "quit",
    "exit",
    "set",
    "reset",
    "unset",
    "isset",
    "isnotset",
    "timeset",
    "istimeset",
    "exec",
    "system",
    "variable"
};

typedef struct hash_struct /* hash table element, the required elements are struct hash_struct *previous; and struct hash_struct *next; other appropriate elements may be added, consistent with the capabilities of the functions in HASHTABLE */
{
    struct hash_struct *previous; /* reference to next element in hash's doubly, circular linked list */
    void *hash_key; /* reference to element's key */
    time_t time_set; /* time of last set */
    int command; /* command code, one of enum keyword */
    struct hash_struct *next; /* reference to previous element in hash's doubly, circular linked list */
} HASH;

#ifdef __STDC__

typedef struct hashtable_struct /* hash table descriptor */
{
    size_t hash_size; /* size of hash table */
    HASH *table; /* reference to hash table index */
    HASH *(*mkhash) (void *key); /* reference to the function that allocates hash elements */
    int (*cmphash) (void *data, struct hash_struct *element); /* reference to the function that compares element's keys */
    void (*rmhash) (struct hash_struct *element); /* reference to the function that deallocates hash elements */
    int (*comphash) (struct hashtable_struct *hash_table, void *key); /* reference to the function that computes the hash value */
} HASHTABLE;

#else

typedef struct hashtable_struct /* hash table descriptor */
{
    size_t hash_size; /* size of hash table */
    HASH *table; /* reference to hash table */
    HASH *(*mkhash) (); /* reference to the function that allocates data and hash elements */
    int (*cmphash) (); /* reference to the function that compares element's keys */
    void (*rmhash) (); /* reference to the function that deallocates data hash elements */
    int (*comphash) (); /* reference to the function that computes the hash value */
} HASHTABLE;

#endif

static HASH *temp_hash; /* temporary HASH reference in SWAP() */

#define SWAP(x,y) temp_hash = (x);(x) = (y);(y) = temp_hash /* linked list link element manipulator */

#ifdef __STDC__

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, int partial_key, char *BOFmap, char *EOFmap);
static void print_message (int pretval); /* print any error messages */
static void spclose (int err); /* interrupt handler */
static int child_read (int fdd, char *begin, char *end, int infiled); /* child read loop, process all read records from the modem */
static int parent_write (int fdd, int infiled, char *query_name); /* parent write loop, process all write records to the modem */
static int usec_sleep (long usecs); /* short sleep for usecs microseconds */
static int usec_delay (long usecs); /* long sleep for usecs microseconds */
static int bits_out (int fdd, int bits); /* send out one command bit */
static int clock_out (int fdd); /* send a clock, (i.e., both RTS and DTR set) */
static int cm17a_out (int fdd, unsigned char device, unsigned char housecode, int cmd); /* send a command */
static int strtoken (char *string, char *parse_array, char **parse, char *delim); /* parse a record based on sequential delimiters */
static int hash_init (HASHTABLE *hash_table); /* initialize the hash table */
static int hash_insert (HASHTABLE *hash_table, void *key); /* insert a key into the hash table */
static HASH *hash_find (HASHTABLE *hash_table, void *key); /* find a key in the hash table */
static int hash_delete (HASHTABLE *hash_table, void *key); /* delete a key from the hash table */
static void hash_term (HASHTABLE *hash_table); /* remove a hash table */
static int hash_text (HASHTABLE *hash_table, void *key); /* compute the hash value for a text key */
static int cmphash (void *data, HASH *element); /* function to compare a text key with a hash table's element key */
static HASH *mkhash (void *key); /* function to allocate a text hash table element and data */
static void rmhash (HASH *element); /* function to deallocate a text hash table element and data */
static int interpret (int fdd, char *buffer, int *quit_flag, char *query_name); /* interpret a command */
static int write_query (char *buffer, size_t len); /* write a record to the query file */
static int write_outfile (char *outbuf, size_t buflen, char *timestamp); /* write a record to the output file */

#else

static char *binarysearch (); /* binary search the database file */
static void print_message (); /* print any error messages */
static void spclose (); /* interrupt handler */
static int child_read (); /* child read loop, process all read records from the modem */
static int parent_write (); /* parent write loop, process all write records to the modem */
static int usec_sleep (); /* short sleep for usecs microseconds */
static int usec_delay (); /* long sleep for usecs microseconds */
static int bits_out (); /* send out one command bit */
static int clock_out (); /* send a clock, (i.e., both RTS and DTR set) */
static int cm17a_out (); /* send a command */
static int strtoken (); /* parse a record based on sequential delimiters */
static int hash_init (); /* initialize the hash table */
static int hash_insert (); /* insert a key into the hash table */
static HASH *hash_find (); /* find a key in the hash table */
static int hash_delete (); /* delete a key from the hash table */
static void hash_term (); /* remove a hash table */
static int hash_text ();  /* compute the hash value for a text key */
static int cmphash (); /* function to compare a text key with a hash table's element key */
static HASH *mkhash (); /* function to allocate a text hash table element and data */
static void rmhash (); /* function to deallocate a text hash table element and data */
static int interpret (); /* interpret a command */
static int write_query (); /* write a record to the query file */
static int write_outfile (); /* write a record to the output file */

#endif

static HASHTABLE hash_variable = {2729, (HASH *) 0, mkhash, cmphash, rmhash, hash_text}; /* declare the hash table descriptor for text keys*/

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    static char pid_str[BUFLEN]; /* pid string */

    char *device = serial_device, /* default serial port device name */
         *in_name = (char *) 0, /* input file name */
         *out_name = (char *) 0, /* output file name */
         *query_name = (char *) 0, /* query file name */
         *db_name = db, /* database file name */
         *begin, /* begining of memory mapped database file */
         *end; /* end of memory mapped database file */

    int c; /* command line switch */

    struct termios newtio; /* device attributes used by this program */

    struct stat fileinfo; /* structure for file status */

    pid_write = getpid (); /* save the pid of parent write process */

    while ((c = getopt (argc, argv, "d:f:i:l:o:q:Vv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'd': /* request for device name? */

                device = optarg; /* yes, save the device name */
                break;

            case 'f': /* request for database file name? */

                db_name = optarg; /* yes, save the database filename */
                break;

            case 'i': /* request for input file name? */

                in_name = optarg; /* yes, save the input file name */
                break;

            case 'l': /* request for lock file name? */

                lockname = optarg; /* yes, save the lock file name */
                break;

            case 'o': /* request for output file name? */

                out_name = optarg; /* yes, save the output file name */
                break;

            case 'q': /* request for query file name? */

                query_name = optarg; /* yes, save the query file name */
                break;

            case 'V': /* request for verbose diagnostics? */

                verbose = 1; /* set the verbose diagnostics flag, 0 = no, 1 = yes */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = ESIG; /* assume error installing signal handler */

        if (signal (SIGINT, spclose) != SIG_ERR) /* install the interrupt handler for SIGINT = ^C */
        {

            if (signal (SIGTERM, spclose) != SIG_ERR) /* install the interrupt handler for SIGTERM */
            {

                if (signal (SIGCHLD, spclose) != SIG_ERR) /* install the interrupt handler for SIGCHLD */
                {
                    lockfile.l_type = F_WRLCK; /* exclusive write lock the entire file */
                    lockfile.l_whence = (short int) SEEK_SET;
                    lockfile.l_start = (off_t) 0;
                    lockfile.l_len = (off_t) 0;

                    retval = ELOCKF; /* assume error creating lock file */

                    if ((semaphorelock = open (lockname, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file exclusively */
                    {
                        remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
                        retval = NOERROR; /* assume no error */
                    }

                    else if ((semaphorelock = open (lockname, O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file if it exists */
                    {
                        retval = NOERROR; /* assume no error */
                    }

                    if (retval == NOERROR) /* any errors? */
                    {
                        retval = EFCNL; /* assume error locking lock file */

                        if (fcntl (semaphorelock, F_SETLK, &lockfile) == 0) /* exclusive write lock the lock file */
                        {
                            remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
                            retval = EWLCK; /* assume error writing to lock file */

                            if (ftruncate (semaphorelock, 0) == 0) /* truncate the lock file */
                            {
                                (void) sprintf (pid_str, "%10d\n", pid_write); /* the pid string is the PID right justified to 10 characters plus a '\n' EOL */

                                if (write (semaphorelock, pid_str, strlen (pid_str)) >= 0) /* write the serial port dialer string to the serial port */
                                {
                                    retval = EOPEN; /* assume error opening device */

                                    if ((fd = open (device, O_RDWR | O_NOCTTY | O_NDELAY)) >= 0) /* open the serial port device file descriptor */
                                    {
                                        retval = EBLOCK; /* assume error enabling blocking on serial device */

                                        if (fcntl (fd, F_SETFL, O_NONBLOCK) == 0) /* enable read blocking, i.e., read () waits until a character is in the buffer */
                                        {
                                            retval = EFLUSH; /* assume error flushing device attributes */

                                            if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                            {
                                                retval = EGATTR; /* assume error getting device attributes */

                                                if (tcgetattr (fd, &oldtio) == 0) /* save original port settings */
                                                {
                                                    state = 1; /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
                                                    memcpy (&newtio, &oldtio, sizeof (newtio)); /* copy the original device attributes */
                                                    retval = EBAUD; /* assume error setting baud rate */

                                                    if (cfsetispeed (&newtio, BAUDRATE) == 0) /* set the baud rate */
                                                    {

                                                        if (cfsetospeed (&newtio, BAUDRATE) == 0)
                                                        {

                                                            /* raw terminal set up, see "man cfmakeraw(3)" for particulars */

                                                            newtio.c_cflag |= (CLOCAL | CREAD); /* local mode, enable receiver */
                                                            newtio.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
                                                            newtio.c_oflag &= ~OPOST;
                                                            newtio.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
                                                            newtio.c_cflag &= ~(CSIZE|PARENB); /* No parity (8N1) */
                                                            newtio.c_cflag |= CS8;
                                                            newtio.c_cc[VTIME] = 0;
                                                            newtio.c_cc[VMIN] = 1;

                                                            retval = ESATTR; /* assume error setting device attributes */

                                                            if (tcsetattr (fd, TCSANOW, &newtio) == 0) /* set the device attributes */
                                                            {
                                                                retval = EFLUSH; /* assume error flushing device attributes */

                                                                if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                                                {
                                                                    retval = EOOPEN; /* assume error opening output file */

                                                                    if ((outfile = (out_name == (char *) 0) ? outfile : open (out_name, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the output file */
                                                                    {
                                                                        retval = EQOPEN; /* assume error opening query file */

                                                                        if ((queryfile = (query_name == (char *) 0) ? queryfile : open (query_name, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the query file, O_SYNC used since the query file may be a named pipe */
                                                                        {
                                                                            retval = EIOPEN; /* assume error opening file */

                                                                            if ((infile = (in_name == (char *) 0) ? infile : open (in_name, O_RDWR)) >= 0) /* open the input file */
                                                                            {
                                                                                retval = ESTAT; /* assume error in status of database file */

                                                                                if (stat (db_name, &fileinfo) == 0) /* get the status of the database file */
                                                                                {
                                                                                    retval = EDB;  /* assume error in database file */

                                                                                    if ((filesize = (size_t) fileinfo.st_size) > (size_t) 0) /* save the size of database file */
                                                                                    {
                                                                                        retval = ESTAT; /* assume error in status of database file */

                                                                                        if ((dbfile = open (db_name, O_RDONLY)) != -1) /* open the database file */
                                                                                        {
                                                                                            retval = EMAP; /* assume error mapping file */

                                                                                            if ((mapbegin = mmap ((void *) 0, filesize, PROT_READ, MAP_PRIVATE, dbfile, (off_t) 0)) != (caddr_t) 0) /* map the database file */
                                                                                            {
                                                                                                retval = EDB;  /* assume error in database file */

                                                                                                if (mapbegin[filesize - 1] == '\n') /* verify that the last record of the memory mapped database file has a terminating '\n' character as required for binarysearch() */
                                                                                                {
                                                                                                    begin = (char *) mapbegin; /* reference the begining of memory mapped database file */
                                                                                                    end = &mapbegin[filesize - 1]; /* reference the end of memory mapped database file */

                                                                                                    retval = NOERROR; /* assume no error */

                                                                                                    if ((pid_read = fork ()) < 0) /* fork the child read process */
                                                                                                    {
                                                                                                        retval = EFORK; /* assume error forking to child read process */
                                                                                                    }

                                                                                                    else if (pid_read == 0) /* child or parent process, (pid_read is zero in the child process, the child process' PID in the parent)? */
                                                                                                    {

                                                                                                        /* start of child read process; pid_write still contains the PID of the parent write process */

                                                                                                        retval = child_read (fd, begin, end, infile); /* child read loop, process all read records from the modem */

                                                                                                        /* end of the child read process, fall through and exit */
                                                                                                    }

                                                                                                    else /* wasn't child process, must be parent */
                                                                                                    {

                                                                                                        /* start of the parent write process; pid_read contains the PID of the child read process */

                                                                                                        retval = parent_write (fd, infile, query_name); /* parent write loop, process all write records to the modem */

                                                                                                        /* end of the parent write process, fall through and exit */

                                                                                                    }

                                                                                                }

                                                                                            }

                                                                                        }

                                                                                    }

                                                                                }

                                                                            }

                                                                        }

                                                                    }

                                                                }

                                                            }

                                                        }

                                                    }

                                                }

                                            }

                                        }

                                    }

                                }

                            }

                        }

                    }

                }

            }

        }

    }

    spclose (0); /* shutdown */
    exit (-retval); /* exit with the error value, formality: spclose () did the exit */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Recursively binary search a memory area.

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, int partial_key, char *BOFmap, char *EOFmap);

I) Binary search a memory area, referenced by the begin and end
variables, for string. The variable begin references the beginning of
the area to be searched; the variable end references the end of the
area to be searched, and will always reference the '\n' terminating
character of the last record of the memory area, (which is a
requirment); the variable string is the null terminated string to
search for, and has a length represented by the variable string_len.

II) Requirements:

    A) The memory area size must be greater than zero, (this is a
    requirement for mmap(1), also.)

    B) The last character of the memory area must be a '\n' text file
    record terminator, e.g., the last record, (which, minimally, may
    be the only record,) must be terminated.

    C) The records in the memory area must be in lexical sequence,
    (consistent with strcmp() and strncmp().)

III) The general strategy is:

    A) For each iteration in the recursion:

        1) Divide the memory area exactly in two.

        2) Work from this middle, in both directions, to find the end
        of the preceding record, (or beginning of file, whichever
        comes first) and, the end of the current record:

            a) The current record contains the first instance of the
            the terminating character at, or past, the middle.

            b) The current record's beginning is the next character
            after the preceding records's terminating character, or
            the beginning of the memory area.

        3) Compare this record with the string for a match.

        4) If not a match, decide which direction to move for the next
        iteration, and call this function, again.

        5) The recursive binary search terminates when a match is
        found, or when there is no more memory area to search.

    B) Return a reference to the matched area in memory, or null if no
    match.

Returns a reference to the matched string, null if no match.

*/

#ifdef __STDC__

static char *binarysearch (char *begin, char *end, char *string, size_t string_len, int partial_key, char *BOFmap, char *EOFmap)

#else

static char *binarysearch (begin, end, string, string_len, BOFmap, EOFmap)
char *begin;
char *end;
char *string;
size_t string_len;
int partial_key;
char *BOFmap;
char *EOFmap;

#endif

{
    int local_retval; /* local return value, return for string comparisons */

    char *middle, /* reference to the middle between begin and end */
         *BOFmiddle_record, /* reference to the beginning of the record in the middle */
         *EOFmiddle_record, /* reference to the end of the record in the middle */
         *BOFnext_record, /* beginning of next record */
         *EOFprevious_record, /* end of previous record */
         *match = (char *) 0; /* reference to the beginning of the record in the middle if a match is found; null if no match */

    size_t middle_record_len; /* length record in the middle */

/* #define QA */ /* comment out to disable QA/Debuging */

#ifdef QA

    /*

    The Unix dictionary file, "sort -u /usr/local/lib/dict/websters >
    testfile", is a good QA test candidate.

    Also "thisprogram testfile < testfile > out" without QA #defined,
    and "cmp testfile out" is a good QA test candidate, (to verify
    that all words are successfully found,) where thisprogram is the
    name of the program containing this function.

    Test for the pathological cases, too; duplicate records, null
    records, single null records, etc.

    */

    char QAbuf[BUFSIZ], /* QA print buffer */
         *QAbegin, /* beginning of record to be printed */
         *QAend; /* end of record to be printed */

    size_t QAreclen; /* QA record length in print buffer */

#endif

    middle = BOFmiddle_record = EOFmiddle_record = begin + ((end - begin) / 2); /* reference the middle between begin and end */

    while ((BOFmap < BOFmiddle_record) && (*(BOFmiddle_record - 1) != '\n')) /* step backwards to the beginning of the record in the middle */
    {
        BOFmiddle_record --;
    }

    while ((EOFmiddle_record < EOFmap) && (*EOFmiddle_record != '\n')) /* step forwards to the end of the record in the middle */
    {
        EOFmiddle_record ++;
    }

    begin = (BOFmiddle_record < begin) ? BOFmiddle_record : begin; /* beginning of record in the middle outside of begin? */
    end = (end < EOFmiddle_record) ? EOFmiddle_record : end; /* end of record in the middle outside of end? */
    BOFnext_record = (EOFmiddle_record < EOFmap) ? EOFmiddle_record + 1 : EOFmap; /* save the beginning of next record */
    EOFprevious_record = (BOFmap < BOFmiddle_record) ? BOFmiddle_record - 1 : BOFmap; /* end of previous record */
    middle_record_len = EOFmiddle_record - BOFmiddle_record; /* save the length of the record in the middle, minus the terminating '\n' */
    local_retval = strncmp (string, BOFmiddle_record, string_len < middle_record_len ? string_len : middle_record_len); /* compare the record in the middle with the search string */

#ifdef QA

    (void) printf ("%s, ", string); /* print the string being searched for */

    QAreclen = strchr (BOFmiddle_record, '\n') - BOFmiddle_record; /* save the length of the record being searched */
    (void) strncpy (QAbuf, BOFmiddle_record, QAreclen); /* copy the record being searched to the print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the record being searched in the print buffer */
    (void) printf ("%s, ", QAbuf); /* print the copy of the record being searched */

    QAreclen = strchr (begin, '\n') - begin; /* length of first record in begin/end */
    (void) strncpy (QAbuf, begin, QAreclen); /* copy first record in begin/end to QA print buffer */
    QAbuf[QAreclen] = '\0'; /* terminate the copy of the first record in begin/end */
    (void) printf ("(%s - ", QAbuf); /* print the copy of the first record in begin/end */

    QAend = QAbegin = end; /* end of last record in begin/end */

    while ((BOFmap < QAbegin) && (*(QAbegin - 1) != '\n')) /* step backwards to the beginning of the last record in begin/end */
    {
        QAbegin --;
    }

    (void) strncpy (QAbuf, QAbegin, (size_t) (QAend - QAbegin)); /* copy last record in begin/end to QA print buffer */
    QAbuf[QAend - QAbegin] = '\0'; /* terminate the copy of the last record in begin/end */
    (void) printf ("%s)\n", QAbuf); /* print the copy of the last record in begin/end */

#endif

    if (local_retval < 0) /* string < record? */
    {

        if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
        {
            match = binarysearch (begin, EOFprevious_record, string, string_len, partial_key, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
        }

    }

    else if (local_retval > 0) /* string > record? */
    {

        if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
        {
            match = binarysearch (BOFnext_record, end, string, string_len, partial_key, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
        }

    }

    else /* match, string = record? */
    {

        if (string_len < middle_record_len) /* partial key match; string shorter than record? */
        {

            if (begin < BOFmiddle_record) /* another potential match between the begining of the record in the middle and begin? */
            {
                match = binarysearch (begin, EOFprevious_record, string, string_len, partial_key, BOFmap, EOFmap); /* return a reference to any potential match between the begining of the record in the middle and begin */
            }

            if (partial_key == 1) /* partial key searche? */
            {

                /*

                The beginning of this record, (and possibly its tab
                field delimiter,) matches the string, (and possibly
                its trailing tab field delimiter); this match is not
                the first unless the variable match == NULL, (i.e., no
                matches were found between begin and
                EOFprevious_record.)

                */

                if (match == (char *) 0) /* if no matchs prior to EOFprevious_record, this is the first match */
                {
                    match = BOFmiddle_record; /* return a reference to this record */
                }

            }

        }

        else if (string_len > middle_record_len) /* partial key match; string longer than record? */
        {

            if (EOFmiddle_record < end) /* another potential match between the end of the record in the middle and end? */
            {
                match = binarysearch (BOFnext_record, end, string, string_len, partial_key, BOFmap, EOFmap); /* return a reference to any potential match between the end of the record in the middle and end */
            }

        }

        else /* exact match */
        {
            match = BOFmiddle_record; /* return a reference to this record */
        }

    }

    return (match); /* return a reference to the match found between begin and end; null if no match */
}

/*

Initialize the hash table.

static int hash_init (HASHTABLE *hash_table);

I) If this function has not been executed for this HASHTABLE
   structure, allocate hash_table->hash_size many HASH structures for
   the hash table index, initialize each element in the hash table
   index's next and previous references to reference the element
   itself.

II) The required argument is a reference to the hash table descriptor
    structure, hash_table.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int hash_init (HASHTABLE *hash_table)

#else

static int hash_init (hash_table)
HASHTABLE *hash_table;

#endif

{
    int local_retval = EHINIT; /* local return value, assume hash table already initilaized */

    size_t i; /* hash table element counter */

    HASH *hash_ref; /* reference to element in hash table index */

    if (hash_table->table == (HASH *) 0) /* protect from repeated calls by testing if hash table index has been allocated */
    {
        local_retval = EALLOC; /* assume memory error */

        if ((hash_table->table = (HASH *) malloc (hash_table->hash_size * sizeof (HASH))) != (HASH *) 0) /* allocate the hash table index */
        {
            local_retval = NOERROR; /* assume no error */

            for (i = 0; i < hash_table->hash_size; i ++) /* initialize the hash table doubly, circular linked list */
            {
                hash_ref = &hash_table->table[i]; /* reference the element in the hash table index */
                hash_ref->next = hash_ref; /* references its own element */
                hash_ref->previous = hash_ref; /* references its own element */
            }

        }

    }

    return (local_retval); /* return any error */
}

/*

Insert a key into the hash table.

static int hash_insert (HASHTABLE *hash_table, void *key);

I) Allocate the key's hash, insert the key and data at the beginning
   of the doubly linked list for this hash table index element, which
   is implicitly addressed by the key's hash.

II) The required arguments are a reference to the hash table descriptor
    structure, hash_table, and a reference to the key value, data.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int hash_insert (HASHTABLE *hash_table, void *key)

#else

static int hash_insert (hash_table, key)
HASHTABLE *hash_table;
void *key;

#endif

{
    int local_retval = EHMK, /* local return value, assume hash table mkhash () failure */
        hindex; /* index of element in hash table */

    HASH *element, /* reference to new element to be inserted in hash table */
         *next_element, /* reference to next element in hash's doubly, circular linked list */
         *previous_element; /* reference to indexed element in the hash table */

    hindex = (*hash_table->comphash) (hash_table, key); /* get the key's hash */
    previous_element = &hash_table->table[hindex]; /* reference the element in the hash table index */
    next_element = previous_element->next; /* start with the first element in the hash's doubly, circular linked list */

    if ((element = (*hash_table->mkhash) (key)) != (HASH *) 0) /* get the hash element */
    {
        SWAP (previous_element->next, element->next); /* the new element is first in the doubly, circular linked list */
        SWAP (next_element->previous, element->previous);
        local_retval = NOERROR; /* assume no error */
    }

    return (local_retval); /* return any error */
}

/*

Find a key in the hash table.

static HASH *hash_find (HASHTABLE *hash_table, void *key);

I) Calculate the key's hash, search for the key and data, starting at
   the beginning of the doubly linked list for this hash table index
   element, which is implicitly addressed by the key's hash.

II) The required arguments are a reference to the hash table
    descriptor structure, hash_table, and a reference to the key
    value, data.

Returns a reference to the hash table's element referencing the data,
0 if key is not found.

*/

#ifdef __STDC__

static HASH *hash_find (HASHTABLE *hash_table, void *key)

#else

static HASH *hash_find (hash_table, key)
HASHTABLE *hash_table;
void *key;

#endif

{
    int local_retval = EHKEY, /* local return value, assume hash table key not found */
        hindex; /* index of element in hash table */

    HASH *next_element, /* reference to next element in hash index's doubly, circular linked list */
         *head_element, /* reference to head element in hash index's doubly, circular linked list */
         *element = (HASH *) 0; /* return value, assume error */

    hindex = (*hash_table->comphash) (hash_table, key); /* get the key */
    head_element = next_element = &hash_table->table[hindex]; /* reference the head element in hash index's doubly, circular linked list */

    while ((next_element = next_element->next) != head_element) /* while the element in the hash index's doubly, circular linked list is not the list's head */
    {

        if ((*hash_table->cmphash) (key, next_element) == 0) /* element compare with the key? */
        {
            element = next_element; /* yes, found it, return a reference to the element */
            local_retval = NOERROR; /* assume no error */
            break;
        }

    }

    return (element); /* return any error */
}

/*

Delete a key from the hash table.

static int hash_delete (HASHTABLE *hash_table, void *key);

I) Calculate the key's hash, search for the key and data, starting at
   the beginning of the doubly linked list for this hash table index
   element, which is implicitly addressed by the key's hash, if found,
   remove the key's HASH element, and data.

II) The required arguments are a reference to the hash table
    descriptor structure, hash_table, and a reference to the key
    value, data.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __ANSI__

static int hash_delete (HASHTABLE *hash_table,void *key)

#else

static int hash_delete (hash_table, key)
HASHTABLE *hash_table;
void *key;

#endif

{
    int local_retval = EHKEY, /* local return value, assume hash table key not found */
        hindex; /* index of element in hash table */

    HASH *element, /* reference to element to be deleted from hash table */
         *next_element, /* reference to next element in hash index's doubly, circular linked list */
         *previous_element, /* reference to previous element in hash index's doubly, circular linked list */
         *head_element; /* reference to head element in hash index's doubly, circular linked list */

    hindex = (*hash_table->comphash) (hash_table, key); /* get the key */
    head_element = element = &hash_table->table[hindex]; /* reference the head element in hash index's doubly, circular linked list */

    while ((element = element->next) != head_element) /* while the element in the hash index's doubly, circular linked list is not the list's head */
    {

        if ((*hash_table->cmphash) (key, element) == 0) /* element compare with the key? */
        {
            next_element = element->next; /* yes, reference the next element in the hash's doubly, circular linked list */
            previous_element = element->previous; /* reference the previous element in the hash's doubly, circular linked list */
            SWAP (next_element->previous,element->previous); /* delete the element from the hash's doubly, circular linked list */
            SWAP (previous_element->next,element->next);
            (*hash_table->rmhash) (element); /* delete the element, and data referenced by the element */
            local_retval = NOERROR; /* found it, assume no error */
            break;
        }

    }

    return (local_retval); /* return any error */
}

/*

Remove a hash table.

static void hash_term (HASHTABLE *hash_table);

I) For each element in the hash table's index, for each element in the
   doubly linked list referenced by the hash table's index element,
   remove the key's HASH element.

II) The required argument is a reference to the hash table descriptor
    structure, hash_table.

There is no return value from this function.

*/

#ifdef __STDC__

static void hash_term (HASHTABLE *hash_table)

#else

static void hash_term (hash_table)
HASHTABLE *hash_table;

#endif

{
    int local_retval = NOERROR; /* local return value, assume no error */

    size_t i;

    HASH *element, /* reference to element being deleted from hash index's doubly, circular linked list */
         *next_element, /* reference to next element in hash index's doubly, circular linked list */
         *head_element; /* reference to head element in hash index's doubly, circular linked list */

    for (i = 0; i < hash_table->hash_size && local_retval == NOERROR; i ++) /* for each of the hash table's index's (while there are no errors) */
    {
        head_element = &hash_table->table[i]; /* reference the head element in hash index's doubly, circular linked list */
        next_element = head_element->next; /* start with the first element in the hash index's doubly, circular linked list */

        while (next_element != head_element) /* while the element in the hash's doubly, circular linked list is not the list's head */
        {
            element = next_element; /* reference the element to be deleted from the doubly, circular linked list */
            next_element = next_element->next; /* reference next element in the hash's doubly, circular linked list */
            SWAP (next_element->previous,element->previous); /* delete the element from the hash's doubly, circular linked list */
            SWAP (head_element->next,element->next);
            (*hash_table->rmhash) (element); /* delete the element, and data referenced by the element */
        }

    }

    free (hash_table->table); /* no, free the hash table index */
}

/*

Compute the hash value for a text key.

static int hash_text (HASHTABLE *hash_table, void *key);

I) The routine requires a long to be 32 bits, which represents some
   portability issues.

II) The required arguments are a reference to the hash table
    descriptor structure, hash_table, and a reference to the key
    value, key.

Returns the hash value for the key.

*/

#ifdef __STDC__

static int hash_text (HASHTABLE *hash_table, void *key)

#else

static int hash_text (hash_table, key)
HASHTABLE *hash_table;
void *key;

#endif

{
    char *char_ref; /* byte reference into data */

    unsigned long num,
                  hash_num = (unsigned long) 0; /* hash number is initially zero */

    for (char_ref = key;*char_ref != (char) 0;char_ref ++) /* for each character in the key */
    {
        hash_num = (hash_num << 4) + (unsigned long) (((int) (*char_ref) < (int) 0) ? (long) (-(*char_ref)) : (long) (*char_ref)); /* multiply the hash by 16, and add the absolute value of the character */

        if ((num = hash_num & 0xf0000000) != (unsigned long) 0) /* any high order bits in bit positions 24 - 28 ? */
        {
            hash_num = hash_num ^ (num >> 24); /* yes, divide by 16777216, and exor on itself */
            hash_num = hash_num ^ num; /* reset the high order bits to 0 */
        }

    }

    return ((int) (hash_num % (unsigned long) hash_table->hash_size)); /* return the hash number */
}

/*

Function that compares a text key with a hash table element key.

static int cmphash (void *data, HASH *element);

(note that the data reference in HASH is a void reference, and
requires a cast to the appropriate data type)

returns 0 if data and the key associated with the element key are
equal

returns non-zero if data and the key associated with the element are
not equal

*/

#ifdef __STDC__

static int cmphash (void *data, HASH *element)

#else

static int cmphash (data, element)
void *data;
HASH *element;

#endif

{
    return (strcmp (data, ((char *) element->hash_key))); /* return the value of the comparison */
}

/*

Function that allocates and loads a hash table element into a hash table.

static HASH *mkhash (void *key);

returns a reference to the element constructed if successful

returns NULL if no element was constructed

Note: local_retval, is set to EALLOC if a memory allocation error
occured

*/

#ifdef __STDC__

static HASH *mkhash (void *key)

#else

static HASH *mkhash (key)
void *key;

#endif

{
    int local_retval = EALLOC; /* local return value, assume error allocating memory */

    void *key_ref; /* reference to key */

    HASH *element = (HASH *) 0; /* reference to the hash table element */

    if ((key_ref = (char *) malloc ((1 + strlen (key)) * sizeof (char))) != (char *) 0) /* allocate area for the key */
    {

        if ((element = (HASH *) malloc (sizeof (HASH))) != (HASH *) 0) /* allocate the hash table element */
        {
            element->next = element; /* all of the elements link references reference the element itself */
            element->previous = element;
            element->hash_key = key_ref; /* reference to element's key */
            (void) strcpy (key_ref, key); /* save the hash table element's key as its data */
            element->time_set = time ((time_t *) 0); /* time of last set, there is no error return if the argument is null */
            element->command = (int) VARIABLE; /* default command is VARIABLE */
            local_retval = NOERROR; /* assume no error */
        }

    }

    return (element); /* return a reference to the hash table element */
}

/*

Function to deallocate a text hash table element.

static void rmhash (HASH *element);

returns nothing

*/

#ifdef __STDC__

static void rmhash (HASH *element)

#else

static void rmhash (element)
HASH *element;

#endif
{
    free (element->hash_key); /* free the key's area allocated in mkhash () */
    free (element); /* free the hash table element allocated in mkhash () */
}

/*

Print any error messages.

static void print_message (int pretval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer pretval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer pretval:

        1) If pretval is zero, print nothing-normal/successful program
           exit.

        2) Else if pretval is unity, print help.

        3) Else pretval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int pretval)

#else

static void print_message (pretval)
int pretval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (pretval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[pretval]);
            break;

    }

}

/*

Interrupt handler and shutdown.

static void spclose (int err);

I) Prerequisites:

    A) All file opening and serial port initialization is done by the
       parent write process, (excepting the log file, which opened and
       closed dynamically by the child read process-and is normally
       closed,)

        1) The parent write process should close all files, (except
           the log file,) and restore the serial port.

        2) The child read process should close the log file.

    B) Called with SIGINT = ^C, or, SIGTERM, or, SIGCHLD interrupts,
       or, with a negative argument representing an internal error,
       or, an argument of zero for no errors or interrupts.

    C) There are "edge" issues, (i.e., a flag is tested to see if
       something needs shutting down, and the something is shut down,
       but the program is interrupted-perhaps by SIGINT = ^C
       bashing-before the flag gets reset,) There are no error message
       printed in the shut down section of this routine-everything is
       shut down, and at the end of this routine, the flags are
       queried for a successful shutdown, and shutdown errors, if any,
       printed. The idea being that if an "edge" issue occurred, the
       shut down would occur twice, (presumably giving an error the
       second time, which would not be printed, but as long as the
       flag is reset, no error would be printed.)

    D) The kill of the child read by the parent write process, (and
       vice verse,) is complex:

        1) static passed = reset
        2) if parent
            a) if err is SIGCHLD
                i) wait for the child read process to die
            b) if passed = reset
                i) reset all things in the parent process
                ii) print any errors
                iii) set passed
                iv) if err not SIGCHLD and child read process active,
                    kill child read process
        3) else if child
            a) if passed = reset
                i) close the log file
                ii) print any errors
                iii) set passed
                iv) kill parent
        4) restore all signals

II) Data structures:

    A) if the pid is equal to pid_write then the parent write process
       is executing this routine, else, it is the child read process.

    B) The parent write and child read process have a separate copy of
       the int passed flag-which means a complete pass has been made
       through this routine for the client or parent.

III) Function execution:

    A) Upon entry:

        1) If the PID of the process is the parent write process:

            a) Shut down all devices opened by the parent write
               process.

            b) Print any pending errors that occurred before the
               program entered this routine.

            c) Print any errors after the shutdown is complete.

            d) If the call to this routine was not caused by a dead
               child read process, and the child process is actve,
               kill the child read process.

        2) If the PID of the process is the child read process:

            a) Shut down all devices opened by the child read process.

            b) Print any pending errors that occurred before the
               program entered this routine.

            c) Print any errors after the shutdown is complete.

            d) Kill the parent write process.

        3) Restore the signal handlers for SIGINT = ^C and SIGTERM.

There is no return from this routine. The routine exits with a value
of zero if no errors or interrupts occurred, or, the value of the
interrupt, (a positive number,) if an interrupt occurred, or, with a
negative error number if an an error occurred.

*/

#ifdef __STDC__

static void spclose (int err)

#else

static void spclose (err)
int err;

#endif

{
    static int passed = 0; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

    int exit_value,
        local_retval = NOERROR; /* local return value, assume no error shutting things down */

    if (pid_write == getpid ()) /* parent write process? */
    {

        if (err == SIGCHLD) /* child read process died? */
        {
            (void) wait ((int *) 0); /* free the child read process */
        }

        if (passed == 0) /* passed this routine, 0 = no, 1 = yes? */
        {

            if (state == 1) /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
            {
                (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */

                if (tcsetattr (fd, TCSANOW, &oldtio) == 0) /* reset the device attributes */
                {
                    state = 0; /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
                    (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */
                }

            }

            if (fd > 0) /* serial port device file descriptor successfully opened? */
            {
                (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */

                if (close (fd) == 0) /* yes, close the device */
                {
                    fd = -1; /* reset the serial port device file descriptor */
                }

            }

            if (mapbegin > (caddr_t) 0) /* reference to beginning of mmap'ed database file set? */
            {

                if (munmap ((void *) mapbegin, filesize) == 0) /* unmap the database file */
                {
                    mapbegin = (caddr_t) 0; /* reset the reference to beginning of mmap'ed file */
                }

            }

            if (dbfile > 0) /* database file descriptor successfully opened? */
            {

                if (close (dbfile) == 0) /* yes, close the database file */
                {
                    dbfile = -1; /* reset the database file descriptor */
                }

            }

            if (infile != STDIN_FILENO) /* input file open? */
            {

                if (close (infile) == 0) /* close the input file */
                {
                    infile = STDIN_FILENO; /* reset the input file descriptor */
                }

            }

            if (queryfile != STDERR_FILENO) /* query file open? */
            {

                if (close (queryfile) == 0) /* close the query file */
                {
                    queryfile = STDERR_FILENO; /* reset the query file descriptor */
                }

            }

            if (outfile != STDOUT_FILENO) /* output file open? */
            {

                if (close (outfile) == 0) /* close the output file */
                {
                    outfile = STDOUT_FILENO; /* reset the output file descriptor */
                }

            }

            if (semaphorelock >= 0) /* lock file open? */
            {
                lockfile.l_type = F_UNLCK; /* unlock the file */
                (void) fcntl (semaphorelock, F_SETLKW, &lockfile); /* unlock the lock file */

                if (close (semaphorelock) == 0) /* close the lock file */
                {
                    semaphorelock = -1; /* reset the lock file file descriptor */

                    if (remove_lockfile == 1) /* remove lock file, 0 = no, 1 = yes, set? */
                    {

                        if (unlink (lockname) == 0) /* remove the lock file */
                        {
                            remove_lockfile = 0; /* remove lock file, 0 = no, 1 = yes */
                        }

                    }

                }

            }

            if (hash_variable.table != (HASH *) 0) /* keyword hash table initialized? */
            {
                hash_term (&hash_variable); /* terminate the hash table */
            }

            print_message (-retval); /* print any error messages from the program, (there can only be zero or one) */

            /* print any errors shutting things down */

            if (state != 0) /* state of serial port still setup? */
            {print_message (-ERATTR); local_retval = ERATTR;} /* print the error message, set the local error shutting things down */

            if (semaphorelock != -1) /* lock file open? */
            {print_message (-ELOCKR); local_retval = ELOCKR;} /* print the error message, set the local error shutting things down */

            /* probably don't care about these-exit () will handle them */

            if (fd != -1) /* serial port device file descriptor still open? */
            {print_message (-ECLOSE); local_retval = ECLOSE;} /* print the error message, set the local error shutting things down */

            if (mapbegin > (caddr_t) 0) /* reference to beginning of mmap'ed database file still set? */
            {print_message (-EUNMAP); local_retval = EUNMAP;} /* print the error message, set the local error shutting things down */

            if (dbfile != -1) /* database file descriptor still open? */
            {print_message (-EDBCLOSE); local_retval = EDBCLOSE;} /* print the error message, set the local error shutting things down */

            if (infile != STDIN_FILENO) /* input file open? */
            {print_message (-EICLOSE); local_retval = EICLOSE;} /* print the error message, set the local error shutting things down */

            if (outfile != STDOUT_FILENO) /* output file open? */
            {print_message (-EOCLOSE); local_retval = EOCLOSE;} /* print the error message, set the local error shutting things down */

            if (queryfile != STDERR_FILENO) /* query file open? */
            {print_message (-EQCLOSE); local_retval = EQCLOSE;} /* print the error message, set the local error shutting things down */

            passed = 1; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

            if (err != SIGCHLD) /* child read process died? */
            {

                if (pid_read != (pid_t) 0) /* child read process forked? */
                {
                    (void) kill (pid_read, SIGTERM); /* kill the child read process */
                }

            }

        }

    }

    else /* no, child read process */
    {

        if (passed == 0) /* passed this routine, 0 = no, 1 = yes? */
        {
            print_message (-retval); /* print any error messages from the program, (there can only be zero or one) */
            passed = 1; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

            if (pid_write != (pid_t) 0) /* parent write process active? */
            {
                (void) kill (pid_write, SIGTERM); /* kill the parent write process */
            }

        }

    }

    /* probably don't care about these-exit () will handle them */

    (void) signal (SIGCHLD, SIG_DFL); /* restore default interrupt for SIGCHLD */
    (void) signal (SIGTERM, SIG_DFL); /* restore default interrupt for SIGTERM */
    (void) signal (SIGINT, SIG_DFL); /* restore default for SIGINT = ^C interrupt */

    exit_value = err; /* assume routine called externally, i.e., SIGINT == ^C or SIGTERM */

    if (exit_value == 0) /* no? */
    {
        exit_value = retval; /* assume error in main body of program, before this routine */

        if (retval == NOERROR) /* no? */
        {
            exit_value = local_retval; /* assume error shutting things down */
        }

    }

    exit (exit_value); /* exit with the value of the interrupt, the value of an error in the main body of this program, the value of an error in this routine, or zero for no error */
}

/*

CM17A error checking:

*/

#define MAX_CMD 7

/*

CM17A command table:

*/

static char cmd_table[] =
{
    (char) 0x00, /* on */
    (char) 0x20, /* off */
    (char) 0x98, /* dim */
    (char) 0x88, /* bright */
    (char) 0x80, /* all off */
    (char) 0x91, /* all on */
    (char) 0x84, /* all lamps off */
    (char) 0x94 /* all lamps on */
};

/*

CM17A Housecode letter part of a device address:

*/

static char housecode_table[] =
{
    0x06, /* A */
    0x07, /* B */
    0x04, /* C */
    0x05, /* D */
    0x08, /* E */
    0x09, /* F */
    0x0a, /* G */
    0x0b, /* H */
    0x0e, /* I */
    0x0f, /* J */
    0x0c, /* K */
    0x0d, /* L */
    0x00, /* M */
    0x01, /* N */
    0x02, /* O */
    0x03 /* P */
};

/*

CM17A unit part of device address:

*/

static char device_table[][2] =
{
    {0x00, 0x00}, /* 1 */
    {0x00, 0x10}, /* 2 */
    {0x00, 0x08}, /* 3 */
    {0x00, 0x18}, /* 4 */
    {0x00, 0x40}, /* 5 */
    {0x00, 0x50}, /* 6 */
    {0x00, 0x48}, /* 7 */
    {0x00, 0x58}, /* 8 */
    {0x04, 0x00}, /* 9 */
    {0x04, 0x10}, /* 10 */
    {0x04, 0x08}, /* 11 */
    {0x04, 0x18}, /* 12 */
    {0x04, 0x40}, /* 13 */
    {0x04, 0x50}, /* 14 */
    {0x04, 0x48}, /* 15 */
    {0x04, 0x58} /* 16 */
};

/*

Short sleep for usecs microseconds.

static int usec_sleep (long usecs);

I) Uses select(2) to avoid using a spin loop and hogging machine
   resources.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int usec_sleep (long usecs)

#else

static int usec_sleep (usecs)
long usecs;

#endif

{
    int local_retval = ETSEL; /* local return value, assume error select timer */

    struct timeval sleeptime; /* time stucture */

    sleeptime.tv_sec = usecs / 1000000; /* seconds */
    sleeptime.tv_usec = usecs % 1000000; /* microseconds */

    if (select (0, (fd_set *) 0, (fd_set *) 0, (fd_set *) 0, &sleeptime) >= 0) /* sleep for the specified time */
    {
        local_retval = NOERROR; /* assume no error */
    }

    return (local_retval); /* return any errors */
}

/*

Long sleep for usecs microseconds.

static int usec_delay (long usecs);

I) Spins through gettimeofday(2) to avoid using a spin loop and
   hogging machine resources.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int usec_delay (long usecs)

#else

static int usec_delay (usecs)
long usecs;

#endif

{
    int local_retval = ETOD; /* local return value, assume error time of day */

    struct timeval currtime, /* current time structure */
        endtime; /* end time structure */

    if (gettimeofday (&endtime, (struct timezone *) 0) >= 0) /* get the current time */
    {
        endtime.tv_usec += usecs; /* set the timer for the specified number of microseconds */

        if (endtime.tv_usec > 1000000) /* more than one second? */
        {
            endtime.tv_sec++; /* increment by the number of seconds */
            endtime.tv_usec -= 1000000; /* and decrease the microseconds by the number of seconds */
        }

        local_retval = NOERROR; /* assume no error */

        do /* spin while time runs down */
        {
            if (gettimeofday (&currtime, (struct timezone *) 0) < 0) /* get the current time */
            {
                local_retval = ETOD; /* assume error time of day */
                break;
            }

        }
        while (timercmp (&endtime, &currtime, >));

    }


    return (local_retval); /* return any errors */
}

/*

Send out one command bit.

static int bits_out (int fdd, int bits);

I) Set RTS or DTR, (but only one,) depending on value.

    A) Assume that both RTS and DTR are high, clearing the one that
       should not be set for this bit.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int bits_out (int fdd, int bits)

#else

static int bits_out (fdd, bits)
int fdd;
int bits;

#endif

{
    int local_retval = EIOTCL, /* local return value, assume error iotcl */
        out = (bits) ? TIOCM_DTR:TIOCM_RTS; /* RTS or DTR? */

    /* Set RTS, DTR to desired settings */

    if (ioctl(fdd, TIOCMBIC, &out) >= 0) /* Set RTS, DTR to desired settings */
    {
        local_retval = usec_delay (InterBitDelay); /* wait a delay */
    }

    return (local_retval); /* return any errors */
}

/*

Send a clock, (i.e., both RTS and DTR set.)

static int clock_out (int fdd);

I) Set RTS and DTR.

    A) Used before and after a command, (long pulse,) and between
       command bits, (short pulse.)

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int clock_out (int fdd)

#else

static int clock_out (fdd)
int fdd;

#endif

{
    int local_retval = EIOTCL, /* local return value, assume error iotcl */
        out = TIOCM_DTR | TIOCM_RTS; /* RTS or DTR? */

    if (ioctl (fdd, TIOCMBIS, &out) >= 0)
    {
        local_retval = usec_delay (InterBitDelay); /* wait the clock delay */
    }

    return (local_retval); /* return any errors */
}

/*

Send a command.

static int cm17a_out (int fdd, unsigned char device, unsigned char housecode, int cmd);

I) Send a command.

    A) The beginning two bytes, (0xd5, 0xaa,) and the and end single
       byte, (0xad,) are the same for all commands, only the two
       middle bytes need to be constructed.

    B) The serial device descriptor should be open.

    C) The housecode is between A and P, inclusive.

    D) The device is between 1 and 16, inclusive.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int cm17a_out (int fdd, unsigned char device, unsigned char housecode, int cmd) /* send a command */

#else

static int cm17a_out (fdd, device, housecode, cmd)
int fdd;
unsigned char unit;
int cmd;

#endif

{
    unsigned char cmd_seq[5] = { 0xd5, 0xaa, 0x00, 0x00, 0xad },
                  byte;

    int local_retval = EIOTCL, /* local return value, assume error iotcl */
        i, /* bit counter */
        j, /* byte counter */
        unit = device,
        out,
        serial_state;

    if ((cmd == DIM) || (cmd == BRIGHT)) /* numeric part of device address must must be 0 for dim or bright */
    {
        unit &= 0xf0;
    }

    if (ioctl (fdd, TIOCMGET, &serial_state) >= 0) /* save state of serial register */
    {
        serial_state &= (TIOCM_DTR | TIOCM_RTS); /* save state of lines */
        serial_state ^= (TIOCM_DTR | TIOCM_RTS); /* both RTS and DTR will be high when finished, remember the state of both */
        local_retval = ECMD; /* assume error command */

        if ((cmd <= MAX_CMD) && (cmd >= 0))
        {
            cmd_seq[2] |= housecode_table[housecode] << 4 | device_table[unit][0]; /* load the middle two bytes of the command sequence */
            cmd_seq[3] |= device_table[unit][1] | cmd_table[cmd];

            if ((local_retval = clock_out (fdd)) >= 0) /* set clock lines and wait for receiver to be ready */
            {

                if ((local_retval = usec_sleep (PreCmdDelay)) >= 0) /* wait the delay */
                {

                    for (j = 0; (j < 5) && (local_retval == NOERROR); j ++) /* for each of the 5 bytes in the command */
                    {
                        byte = cmd_seq[j]; /* get the byte */

                        for (i = 0; (i < 8) && (local_retval == NOERROR); i ++) /* for each bit in the byte */
                        {
                            out = (byte & 0x80) ? 1 : 0; /* get the bit */
                            byte <<= 1;

                            if ((local_retval = bits_out (fdd, out)) == NOERROR)
                            {
                                local_retval = clock_out (fdd);
                            }

                        }

                    }

                    if (local_retval == NOERROR) /* any errors */
                    {

                        if ((local_retval = clock_out (fdd)) >= 0) /* close with a clock, and a wait */
                        {

                            if ((local_retval = usec_sleep (PostCmdDelay)) == NOERROR) /* wait the delay */
                            {
                                local_retval = EIOTCL; /* assume error iotcl */

                                if (ioctl (fdd, TIOCMBIC, &serial_state) >= 0) /* reset the serial state */
                                {
                                    local_retval = NOERROR; /* assume no error */
                                }

                            }

                        }

                    }

                }

            }

        }

    }

    return (local_retval); /* return any errors */
}

/*

Parse a record based on sequential delimiters.

static int strtoken (char *string, char *parse_array, char **parse, char *delim);

Parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field.

Returns the number of fields parsed.

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string,
*parse_array,
**parse,
*delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}

/*

Child read loop, process all records from the modem.

static int child_read (int fdd, char *begin, char *end, int infiled);

I) Data structure:

    A) The serial port device file descriptor, fdd, should be open, and
       available.

    B) The memory mapped database file should be open, and begin
       should reference the beginning of the mapped file, and end
       should reference the end of the file.

    C) The input file descriptor, infiled, should be open and
       available; it is probably stdin, (but can be a named pipe,
       etc.) This file descriptor "owned" by the parent write process,
       but operated on by this child_read process' duplicate copy of
       the program.

    D) The probability that a sequence of transmissions, (to a maximum
       of five of them-which is what X10 devices send,) is valid as a
       function of the number of sequential transmissions, v, (as
       defined in the variable int validation,) with a probability p
       of being corrupt, (a transaction is defined as five identical
       transmissions of X10 data):

        1) The chances that a transmission, (one of the five,) is
           corrupted, and results in a specific transmission, (i.e.,
           something else was sent, and due to corruption resulted in
           say, "a5 on,") is ((1 / (2^16)) * p) if there is an even
           probability, (a valid assumption unless it is known
           otherwise,) of it being any value, (i.e., the same
           probability of being "a5 on" as, say, "b2 off" even though
           "p15 dim" was sent.)

           For v many in sequence, the probability is ((1 / (2^16)) *
           p)^v.

        2) The chances, (say, "a5 on,") of any transmission being
           corrupted, (and resulting in something else,) is p.

           For v many in sequence, the probability is p^v

        3) The probability that a transmission is not corrupted, (due
           to either,) is:

           (1 - ((1 / (2^16)) * p)^v) * (1 - (p^v))

           If p = 0.50, (i.e., a 50% chance, or one in two, of any
           transmission of the five transmissions in a transaction
           being corrupted, which corresponds to, about a 6 dB (S + N)
           / N):

            a) If v = 1, the probability of the transmission being
               detected correctly is 0.499996185302734375, or an error
               rate of about 1 in 2, for each of the five
               transmissions in the transaction; the combined
               probability for all 5 would be 1 - (1 -
               0.499996185302734375)^5 = 0.96874880788891445937, or
               about 3 errors in 100 transactions of 5 transmissions.

            b) If v = 2, the probability of the transmissions being
               detected correctly is 0.74999999995634425431, or an
               error rate of about 3 in 4, for each of the 4 possible
               combinations in the five transmissions in the
               transaction; the combined probability for all 4 would
               be 1 - (1 - 0.74999999995634425431)^4 =
               0.99609374999727151589, or about 1 error in 250
               transactions of 5 transmissions.

            c) If v = 3, the probability of the transmission being
               detected correctly is 0.87499999999999961142, or an
               error rate of about 1 in 8, for each of the 3 possible
               combinations in the five transmissions in the
               transaction; the combined probability for all three
               would be 1 - (1 - 0.87499999999999961142)^3 =
               0.99804687499999998179, or about 1 error in 500
               transactions of 5 transmissions.

            d) If v = 4, the probability of the transmission being
               detected correctly is 0.93750000000000000000, or an
               error rate of about 1 in 16, for each of the 2 possible
               combinations in the five transmissions in the
               transaction; the combined probability for both would be
               1 - (1 - 0.93750000000000000000)^2 = 0.99609375, or
               about 1 error in 256 transactions of 5 transmissions.

           The optimum number of sequential identical transmissions in
           a transaction of 5 identical transmissions is 3 for a
           validation of the data, as set by the variable int
           validation.  (An error rate of 1 in 500 for a (S + N) / N
           of 6 dB is probably acceptable for the task at hand-this is
           about the limit of 50% intelligibility for a short wave/ssb
           radio at 6 dB (S + N) / N, which by comparison is rather
           impressive for such a simple ECC scheme of counting like
           sequential transmissions in a transaction.) There are other
           alternative schemes, although more complicated to
           implement, which may provide superior ECC-perhaps like a
           voting scheme, (odd two out, if they are not the same,
           etc.)

           Note that there exists only 3 possible sequences, (TTTFF,
           FTTTF, FFTTT,) in the 5 transmissions with errors for a
           positive detection-and they are mutually exclusive, (unlike
           v = 2, where, for example, TTFTT would trigger a double
           detection.)

           Note that for a transmission to be included in the sequence
           of a transaction, it must be identical to the last
           transmission, and be received within 2 seconds of the last
           transmission, (the transmission time is 200 ms., which
           could span a "time(2) granularity" second "edge"; 200
           ms. can be determined from the X10 protocol documentation,
           and presumably used by X10 devices.) If a transmission is
           corrupt, and the corruption does not evaluate to an X10
           device listed in the database, it will be skipped, i.e.,
           TFTFT would evaluate to a valid transaction, even though
           the valid transmissions were not sequential, if the
           corruptions resulted in non-X10 codes not in the database,
           (i.e., similar to a "voting process.") The chances of a
           random transmission being found in the database is 644 /
           65536 = 0.0099, since there are, currently, 644 records in
           the database, (as per the X10 protocol specification,) and
           16 bits of data, making the probabilities outlined, above,
           pessimistic. (Note that the MR26A does some filtering, for
           example transmissions received with corrupt header and/or
           footer data will be rejected, but collisions-transmissions
           from different X10 devices transmitted at nearly the same
           time-can corrupt the two interior data fields of the
           transmission.)

           To evaluate information quality received from X10 devices,
           enable verbose diagnostics, (the -V argument.) All
           transmissions of all transactions will then be written to
           the output file.

II) Function execution:

    A) This routine is a "spin loop" that waits for a a character from
       the serial device using select(2), a signal, or, an error to
       terminate the program, while reading characters from the serial
       device file descriptor, fdd.

    B) Input characters from the serial device are written to the
       output file.

    C) If the input characters constitute a valid X10 transaction of
       int validation many sequential transmissions, any commands
       listed in the database entry for the transaction are executed,
       by locking the input file, and writing the commands to
       parent_write ().

       For a transaction to be considered valid, int validation many
       identical transmissions must be received, each within 2 seconds
       of each other, (i.e., in a "re-triggerable one-shot" fashion.)

III) The interaction between the child_read () and parent_write ()
     functions:

    A) When an X10 command is received by the parent_write ()
       function, it is interpreted by the interpret () function, which
       sends the command to the CM17A. Note that for parent_write ()
       to receive the command, the input file, (usually stdin, but can
       be a named pipe,) is locked.

    B) The CM17A sends the command to the MR26A, (and other x10
       devices,) interpret () returns, and the input file unlocked.

    C) The child_read () function receives the command, (via the
       MR26A,) locks the input file, and sends the received data to
       parent_write (), which processes, (perhaps setting variables,
       etc., via the database file,) in response to the command. The
       lock is then removed from the input file.

    D) Note that the program uses the RF link to communicate between
       the parent_write () and child_read () processes-the concurrency
       issues are addressed through locking the input file, and the
       serial port buffers, (which may be required to fifo data from
       the MR26A while the input file is locked by another process,
       and sending X10 commands to the CM17A.)

    E) In this way, the program will not set any variables, (as per
       the database file,) until they have been successfully
       transmitted by the CM17A, and received by the MR26A as an ACK,
       (whether the transmission was received by other X10 devices is
       another issue.)

    F) Note, also, that the -V, (verbose diagnostics,) prints all
       transmission receptions, (there should be five receptions per
       transaction-the same signal data, repeated five times-that's
       the X10 specification,) of X10 data from the MR26A, followed by
       the record looked up in the database, and any transmissions to
       the CM17A involved during interpretation of the record by
       interpret (). Both the parent_write () and child_read ()
       processes write information to the output file, serially, and
       concurrency control of the writes is controlled by the lock on
       the input file-each reception of the five identical data
       transmissions results in locking, and unlocking the input file.

       For example, suppose "a5 on" has the following record in the
       database:

           d5 aa 60 40 ad a5 on the_switch a8 on
           .
           .
           .
           d5 aa 60 58 ad  a8 on the_other_switch set wxyz

       i.e., when a5 is turned on, (by an X10 device, or by external
       input to the program via the input file,) a8 should be turned
       on, and the output file would have the entry:

           Apr 27 13:59:46 2006 d5 aa 60 40 ad a5 on the_switch a8 on
           Apr 27 13:59:46 2006 a8 on
           Apr 27 13:59:46 2006 d5 aa 60 40 ad a5 on the_switch a8 on
           Apr 27 13:59:46 2006 d5 aa 60 40 ad a5 on the_switch a8 on
           Apr 27 13:59:46 2006 d5 aa 60 40 ad a5 on the_switch a8 on
           Apr 27 13:59:46 2006 d5 aa 60 40 ad a5 on the_switch a8 on
           Apr 27 13:59:46 2006 a8 on d5 aa 60 58 ad set wxyz
           Apr 27 13:59:46 2006 a8 on d5 aa 60 58 ad set wxyz
           Apr 27 13:59:46 2006 a8 on d5 aa 60 58 ad set wxyz
           Apr 27 13:59:46 2006 a8 on d5 aa 60 58 ad set wxyz
           Apr 27 13:59:46 2006 a8 on d5 aa 60 58 ad set wxyz

       showing that "a5 on" was received from the MR26A, (and the
       first of five such transmissions was decoded,) the record
       corresponding to "a5 on" looked up in the database, which
       resulted in "a8 on" being sent from the CM17A, (and all five
       transmissions were received by the MR26A.) The value of int
       validate was 1, (detect on the first of 5 transmissions in the
       transaction "a5 on.")

       Note that the function of this program is to function as an
       automated X10 "clicker," (i.e., like an automated HR12A.)

    G) The timing for the CM17A, (from
       http://software.x10.com/pub/manuals/cm17a_protocol.txt):

        1) Serial port timing:

            a) A bit time, (from the serial port,) is at least 0.5
               ms., for logical '0' or '1', plus at least a 0.5
               ms. wait time.

            b) There are 40 serial port bits per transmission:

                i) 16 bit header, ("d5 aa".)

                ii) 16 bit data, ("00 00" through "ff ff".)

                iii) 8 bit footer, ("ad".)

            c) The variable int InterBitDelay controls the bit
               time-both the bit time and wait time are equal to the
               value of InterBitDelay, or a CM17A serial clock time is
               twice the InterBitDelay, per bit.

            d) The variable int PreCmdDelay controls the wait before
               bits are shuffled out at bit time rate, (see below.)

            e) The variable int PostCmdDelay controls the wait after
               bits are shuffled out at bit time rate, (see below.)

        2) Transmission timing, (from
           http://www.printcapture.com/files/X10_RF_Receiver.pdf):

            a) The X10 RF protocol is the same as the NEC IR protocol:

                i) A 9 ms. burst.

                ii) A 4.5 ms. silence to let the AGC settle.

                iii) Two 27 ms. data streams of 16 bits for 54 ms. total.

                iv) 40 ms. of silence.

           for 107.5 ms. for each transmission of a five transmissions
           packet, or a total of 5 * 107.5 ms. = 537.5 ms.

           However, analysis, (from Figure 2,
           http://www.printcapture.com/files/X10_RF_Receiver.pdf,)
           indicates that at least the HR12A sends 6 transmissions per
           packet, and a mystery packet, for a total transmission time
           of 0.9 seconds-which are undocumented. The analysis also
           indicates that the variability in the internal clocks used
           in X10 devices is -5%, to +16%, meaning that the worst case
           transmission time per packet would be 0.9 * 1.16 = 1.044
           seconds.

           Analysis:

            a) The waveform in Figure 2,
               http://www.printcapture.com/files/X10_RF_Receiver.pdf
               is about 1.581395349% slow, (546 / 537.5.)

            b) An extra 6'th transmission is included, adding 107.5
               ms. after correction by a factor of 1.01581395349.

                i) Note that the program counts the like
                   transmissions, and resets its counter at a count of
                   five. If a sixth transmission is present, the sixth
                   will be considered a new transmission, and since
                   there is only a single transmission, it will be
                   dropped, (see the reset of bor_cnt in child_read ()
                   for consequences.)

                ii) It is not clear what effect, if any, this will
                    have with dim/bright commands-they can be
                    sequential transmissions, of one transmission for
                    certain X10 devices.

            c) A space of 84 ms. / 1.01581395349 = 82.693 ms.

            d) Mystery data of 124 ms. / 1.01581395349 = 122.070 ms.

            f) It is not clear, (since its not documented,) whether
               there is supposed to be a space after the mystery
               data-it is assumed that it is not.

            e) Which totals to 849.762 ms., nominal, or 849.762 * 1.16
               = 985.724 ms. maximum.

                i) Note that a transaction can not exceed 985.724 ms.,
                   or a transaction must be concluded in at least two
                   time tick seconds, (see the reset of bor_cnt in
                   child_read () for consequences.)

    H) Power line timing for the RR501/TM751, (from
        http://www.x10.com/pressroom/pressroom_coretechn_press.htm):

        1) Two identical groups:

            a) 11 line cycles each, (2 start bits, 4 house code
               bits, 5 device code bits.)

            b) Separated by 3 line cycles.

           for 25 line cycles, or 25 / 60 = 0.416667 seconds.

           An additional line cycle should be added for RR501/TM751
           synchronization with this program, (the line zero crossing
           comparators may have to wait as much as one complete cycle
           to start the transmission sequence,) or 26 / 60 = 0.433334
           seconds.

    I) MR26A timing:

        1) The reception timing would be the same as transmission
           timing for the CM17A.

        2) The serial port will have a latency of 40 bits, at 9600 BPS,
           for 40 / 9600 = 0.004167 seconds.

    J) Considering a worst case scenario, (a motion
       sensor->MR26A->serial line->computer->clock
       line->CM17A->RR501/TM751->power line->light switch:

        1) Motion sensed.

        2) Transmission latency = 985.724 ms.

        3) Serial latency through MR26A = 4.167 ms.

        4) Transmission detected on third transmission of packet =
           ((107.5 * 3) * 1.16) + 4.167 = 378.267 ms.

            a) The earliest the CM17A can transmit is at 985.724 ms.
               from the beginning of the transmission, (otherwise, the
               response sent by the CM17A will "clobber," or be
               "clobbered" by the initiating packet.)

               Assuming a sensor's transmission starts at time 0, the
               detection of the third transmission could occur as soon
               as the end of the second transmission, (i.e.,
               instantaneous,) or at 107.5 * 2 * 1.16 = 249.4 ms. The
               serial link delay through the MR26A would be 4.167 ms.,
               for a total of 249.4 + 4.167 = 253.567 ms.

               Or:

                   (InterBitDelay * 2 * 40) + PreCmdDelay =
                       985.724 - 253.567 = 732.187 ms. = D1

               Since the recommended minimum is 0.5 ms., a factor
               of 3 will be used, or InterBitDelay = 1.5 ms., making
               PreCmdDelay = 612.187 ms.

        5) CM17A transmission latency = 985.724 ms.

        6) Power line latency of RR501/TM751 = 433.334 ms.

        7) For a total worst case maximum response time of 985.724
           ms. + 985.724 ms. + 433.334 ms. = 2.404782 seconds,
           (849.762 ms. + 849.762 ms. + 433.334 ms. = 2.132858 seconds
           nominal.)

        8) The minimum time between adjacent transmissions to the
           CM17A:

            a) Since the lead time before the CM17A starts
               transmitting is D1 + 4.167 = 732.187 + 4.167 = 736.354
               ms., it could be initiated 736.354 ms. before the end
               of the power line latency; however, this would put the
               start of the next transmission in the current
               transmission.

               This means that the PostCmdDelay has to be at least
               985.724 ms. (Note: 736.354 ms. > 433.334 ms, a
               requirement for this specific solution.)

          Note that this only concerns adjacent, (i.e., sequential,)
          transmissions, (probably a command sequence from the
          database file.) The input file will be locked for the
          duration of this delay, (in addition to the beginning of the
          transaction.) The time between sequential transactions is D1
          + 985.724 + 4.167 = 732.187 + 985.724 + 4.167 = 1.722078
          seconds.

        9) Or the values, determined here:

               INTERBITDELAY 1500
               PRECMDDELAY 612187
               POSTCMDDELAY 985724

           Giving a response time, (sensor-> ... ->switch,) of 2.4
           seconds maximum, (2.1 seconds, nominal,) and a sequential
           latency of 1.7 seconds.

           Compared with a sensor->RR501/TM751->switch time of 985.724
           + 433.334 = 1.419058 seconds, maximum, (849.762 + 416.667 =
           1.266429 seconds , nominal.) Or, inserting the
           MR26A->computer->CM17A link adds about one second to the
           response time, which is an increase of 34 to 48 percent,
           (if the RR501/TM751 and switch require the mystery data to
           conclude the transaction, and there are no other mystery
           issues that would effect timing-and the context change,
           i.e., receive->transmit, time in the X10 devices is
           negligible.)

           Reasonable, conservative assumptions would be to assume the
           mystery data is followed by the same space time as preceded
           the data, or add 82.693 ms. to PreCmdDelay, or PreCmdDelay
           = 612.187 + 82.693 = 694.880 ms., or about 700.000 ms.,
           which would accommodate the context changes, also-an
           increase of about 14%. Adding the same ratio to
           PostCmdDelay, to accommodate context changes in the
           RR501/TM751, results in 1.127152 seconds, and rounding up:

               INTERBITDELAY 1500
               PRECMDDELAY 700000
               POSTCMDDELAY 1200000

           Which is conservative, and would add only 88 ms. to the
           response time, (an increase of 2.5 seconds maximum against
           1.26 seconds nominal, or about double, over the nominal
           sensor->RR501/TM751->switch time.)

        10) The original values used in BottleRocket:

               static int InterBitDelay = 1400;
               static int PreCmdDelay = 300000;
               static int PostCmdDelay = 300000;

           which was found empirically.

    K) A note about the serial cache. While the input file is locked
       by an execution of a command, (either external, or from the
       child_read() process,) serial data is cached, (typically in an
       8KB buffer by the OS.) With a data rate of (40 * 5) / 8 bytes
       per 1.044 seconds = 23.946 bytes / second, or the cache
       can hold 8192 / 23.946 = 342 seconds, (about 6 minutes,)
       of data before data is dropped. (The program can send and
       receive data concurrently, however, when the parent_write ()
       process "owns" the lock on the input file, and the child_read
       () process receives a transmission that initiates a response,
       the response must wait on the parent_write () process lock
       before processing more serial data.)

       Note that the lock is not released until after the
       int PostCmdDelay silence/wait.

    L) There is an undocumented feature-the program will respond to
       the standard X10 command "a dim" or "a dim 5" for comparison
       with X10 remotes. With the -V option results in approximately,
       (within one dim level,) the same DIM/BRIGHT operation as X10
       HR12A dim button. (Apparently, X10 devices require reception of
       all five transmissions of a transaction for validation.)

       Note: for testing, it may be necessary to have an intervening
       address to another device, (i.e., "a6 off,") before testing the
       DIM/BRIGHT on a lamp, (i.e., "a5 dim 3.")

Returns any errors encountered, NOERROR on success.

*/


#ifdef __STDC__

static int child_read (int fdd, char *begin, char *end, int infiled)

#else

static int child_read (fdd, begin, end, infiled)
int fdd;
char *begin;
char *end;
int infiled;

#endif

{
    char buffer[BUFLEN], /* input read buffer */
         outbuf[BUFLEN], /* formatted output buffer */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         token_separators[] = "\t", /* file record field separators */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         *timestamp, /* time stamp */
         *outref = &outbuf[0], /* character reference into formatted output buffer */
         *eor, /* character reference to the end of formatted output buffer */
         *bor, /* character reference to beginning of formatted output buffer */
         *last_bor = (char *) 0; /* character reference to the last beginning of formatted output buffer */

    int local_retval = NOERROR, /* local return value, (so EREAD = "Error reading from device" will not go to spclose () on a SIGINT = ^C or SIGTERM "edge" in this routine,) assume no error */
        fields, /* number of fields in a record */
        bor_cnt = 0, /* count of sequential identical formatted output buffer instances */
        partial_key = 1, /* partial key search, 0 = no, 1 = yes */
        print_flag = 0; /* print flag, 0 = don't print the record, 1 = print the record */

    size_t i, /* character counter */
           len = 0, /* length of the formatted output buffer */
           ilen; /* length of a command from the database file */

    ssize_t cnt; /* count of characters read */

    time_t time_stamp, /* ctime for time stamps */
           last_time_stamp = (time_t) 0; /* last ctime for time stamps */

    struct flock locallock; /* file locking structure */

    fd_set fds; /* select () struct */

    locallock.l_type = F_WRLCK; /* lock/unlock EOF of file for write */
    locallock.l_whence = (short int) SEEK_SET;
    locallock.l_start = (off_t) 0;
    locallock.l_len = (off_t) 0;

    FD_ZERO (&fds); /* null the select () struct */
    FD_SET (fdd, &fds); /* and install it for the serial file descriptor */

    while (local_retval == NOERROR) /* while no error, spin though getting charcters from the serial device, (actually, waiting for a SIGINT = ^C or SIGTERM) */
    {
        local_retval = ESEL; /* assume select error */

        if (select (fdd + 1, &fds, (fd_set *) 0, (fd_set *) 0, (struct timeval *) 0) != -1) /* wait for characters from the serial device */
        {

            if (FD_ISSET (fdd, &fds)) /* data ready? */
            {
                local_retval = EREAD; /* assume error reading from device */

                if ((cnt = read (fdd, buffer, BUFLEN)) > 0) /* read characters from the serial device */
                {
                    local_retval = NOERROR; /* assume no error */

                    for (i = 0; ((ssize_t) i < cnt) && (local_retval == NOERROR); i ++) /* for each character read from the serial device, while there are no errors */
                    {
                        (void) sprintf (outref, "%2.2x %c", (unsigned char) buffer[i], (char) 0); /* format the character as a hex integer */

                        if (strncmp (outref, "ad ", (size_t) 3) == 0) /* 0xad is end of sequence; is it? */
                        {
                            local_retval = ETIME; /* assume error getting time */

                            if ((time_stamp = time ((time_t *) 0)) != -1) /* get the time */
                            {

                                if ((timestamp = ctime (&time_stamp)) != (char *) 0) /* format the local time */
                                {
                                    local_retval = NOERROR; /* assume no error */
                                    outref[2] = '\t'; /* add a tab character for the partial key match to the formatted output buffer */
                                    outref[3] = '\0'; /* terminate the formatted output buffer */
                                    len = strlen (outbuf); /* get the length of the formatted output buffer */
                                    print_flag = 0; /* reset the print flag, 0 = don't print the record, 1 = print the record */

                                    if ((bor = binarysearch (begin, end, outbuf, len, partial_key, begin, end)) != (char *) 0) /* search the file map for the string */
                                    {

                                        if ((bor_cnt > 5) || (bor != last_bor) || (difftime (time_stamp, last_time_stamp) > (double) 1.0)) /* record not part of sequence of 5 transmitted? */
                                        {
                                            bor_cnt = 0; /* reset the count of sequential identical formatted output buffer instances */
                                        }

                                        last_bor = bor; /* save the character reference to the last beginning of formatted output buffer */
                                        last_time_stamp = time_stamp; /* save the last ctime for time stamps */
                                        eor = strchr (bor, '\n'); /* reference the end of formatted output buffer */
                                        len = eor - bor + 1; /* get the length of the formatted output buffer */
                                        (void) strncpy (outbuf, bor, len); /* copy the data from the database into the formatted output buffer */
                                        outbuf[len] = '\0'; /* terminate the formatted output buffer */

                                        if (++ bor_cnt == validation) /* increment the count of sequential identical formatted output buffer instances, is it equal to the validation number? */
                                        {
                                            print_flag = 1; /* set the print flag, 0 = don't print the record, 1 = print the record */

                                            if (infiled != STDIN_FILENO) /* input file descriptor stdin means no commands will be executed MUST BE A NAMED PIPE FOR COMMAND EXECUTION */
                                            {

                                                if ((fields = strtoken (outbuf, parsebuffer, token, token_separators)) > 3) /* parse the output into fields, the forth, if any, is a command, more than 3? */
                                                {

                                                    /*

                                                    There is a command
                                                    in this record,
                                                    (token[3],) which
                                                    will be written
                                                    verbatim to the
                                                    parent_write()
                                                    function for
                                                    execution.

                                                    */

                                                    locallock.l_type = F_WRLCK; /* lock EOF of file for write */
                                                    local_retval = ELLCK; /* assume error locking input file of the parent_write() function*/

                                                    if (fcntl (infiled, F_SETLKW, &locallock) == 0) /* write lock the EOF of the input file of the parent_write() function */
                                                    {
                                                        ilen = strlen (token[3]); /* save the length of a command from the database file */
                                                        local_retval = EIWRITE; /* assume error writing to input file of the parent_write() function */

                                                        if (write (infiled, token[3], ilen) == (ssize_t) ilen) /* write the command to the input file of the parent_write() function */
                                                        {

                                                            if (write (infiled, "\n", (size_t) 1) == (ssize_t) 1) /* write an EOL to the input file of the parent_write() function */
                                                            {
                                                                locallock.l_type = F_UNLCK; /* unlock the file */
                                                                local_retval = ELUCK; /* assume error unlocking file */

                                                                if (fcntl (infiled, F_SETLKW, &locallock) == 0) /* unlock the input file of the parent_write() function */
                                                                {
                                                                    local_retval = NOERROR; /* assume no error */
                                                                }

                                                            }

                                                        }

                                                    }

                                                }

                                            }

                                        }

                                    }

                                    else
                                    {
                                        outref[2] = '\n'; /* terminate the formatted output buffer with an EOL */
                                        outref[3] = '\0'; /* terminate the formatted output buffer */
                                        len = strlen (outbuf); /* get the length of the formatted output buffer */
                                    }

                                    if ((print_flag == 1) || (verbose == 1)) /* print flag, 0 = don't print the record, 1 = print the record, set, or, verbose diagnostics, 0 = no, 1 = yes? */
                                    {
                                        local_retval = write_outfile (outbuf, len, timestamp); /* write the record to the output file */
                                    }

                                    outref = &outbuf[0]; /* reset the character reference into formatted output buffer */
                                }

                            }

                        }

                        else
                        {
                            outref = strchr (outref, '\0'); /* move to the end of the data in the formatted output buffer */
                        }

                    }

                }

                else if ((cnt == 0) && (local_retval == EREAD))
                {

                    /*

                    A read(2) return of zero is valid; it means
                    EOF was encoutered before the end of record
                    sequence, (0xad); loop back and wait for the
                    rest of the record.

                    */

                    local_retval = NOERROR; /* assume no error */
                }

            }

        }

    }

    return (retval = local_retval); /* return any errors */
}

/*

Parent write loop, process all records from the input file, (usually,
stdin,) to the serial device.

static int parent_write (int fdd, int infiled, char *query_name);

I) Data structure:

    A) The serial port device file descriptor, fdd, should be open, and
       available.

    B) The input file descriptor, infiled, should be open and
       available; it is probably stdin, (but can be a named pipe,
       etc.)

    C) The query file name, query_name, should be NULL if stderr is
       used for the query file, or reference an EOS terminated
       character string of the file name.

II) Function execution:

    A) This routine is a "spin loop" that waits for a signal, or, an
       error to terminate the program, while reading characters from
       the input file descriptor, infiled.

    B) Input characters from the input file are formatted/interpreted:

       1) Sent to the serial line.

       2) Variable states are written to the query file.

    C) EOT processing, (for concurrency/flow control and
      interoperability with other processes which write to the input
      file,) is handled as follows:

       1) If an EOT is encountered, the remaining characters after the
          EOT, if any, in the buffer are discarded, and after the
          remaining characters/commands in the buffer are processed,
          an EOT character is written to the query file, as an EOT =
          ^D = end of transmission/transaction marker.

       2) If an EOT is not encountered in a transaction, (which may be
          one record query, or multiple buffers of many record
          queries,) then EOT is not written to the query file-i.e.,
          normal file operations.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int parent_write (int fdd, int infiled, char *query_name)

#else

static int parent_write (fdd, infiled, query_name)
int fdd;
int infiled;
char *query_name;

#endif

{
    char inbuffer[BUFLEN + 1], /* input read buffer */
         outbuffer[BUFLEN + 1], /* output write buffer */
         *ref, /* character reference */
         *sentinel = &outbuffer[BUFLEN], /* reference to sentinel character in output write buffer */
         inchar; /* input character */

    int local_retval = NOERROR, /* local return value, (so ERDIN = "Error reading input file" will not go to spclose () on a SIGINT = ^C or SIGTERM "edge" in this routine,) assume no error */
        number, /* number of command interpreter keywords in hash table */
        quit_flag = 0, /* quit flag, 0 = no, 1 = yes */
        eot_flag = 0, /* end of transaction to be printed to query file, 0 = no, 1 = yes */
        i, /* general counter */
        j; /* general counter */

    ssize_t cnt = 0; /* count of characters read */

    HASH *element;  /* reference to hash table element */

    if ((local_retval = hash_init (&hash_variable)) == NOERROR) /* initialize the hash table */
    {
        number = sizeof (keyword) / sizeof (char *); /* get the number of keywords */

        for (i = 0; (i < number) && (local_retval == NOERROR); i ++) /* for each keyword */
        {
            (void) strcpy (inbuffer, keyword[i]); /* save a copy of the keyword */

            if ((local_retval = hash_insert (&hash_variable, inbuffer)) == NOERROR) /* add a keyword to the keyword hash table */
            {
                local_retval = EHKEY; /* assume hash table key not found */

                if ((element = hash_find (&hash_variable, inbuffer)) != (HASH *) 0) /* find the keyword in the keyword hash table */
                {
                    element-> command = i; /* add the command keyword; it was initialized to VARIABLE as default */
                    local_retval = NOERROR; /* assume no error */
                }

            }

        }

        *sentinel = '\0'; /* sentinel character in output write buffer */
        ref = &outbuffer[0]; /* first character in output write buffer */

        while ((local_retval == NOERROR) && (quit_flag == 0) && ((cnt = read (infiled, inbuffer, BUFLEN)) > 0)) /* read characters from the input file while no errors, or quit request */
        {
            inbuffer[cnt] = '\0'; /* terminate the characters with an EOS for safety, (if the buffer is full, this will overwrite the existing EOS sentinel) */

            for (j = 0; (local_retval == NOERROR) && (quit_flag == 0) && (j < cnt); j ++) /* for each character in the input read buffer */
            {
                local_retval = EWROVFL; /* assume error output write buffer overflow */

                if (ref < sentinel) /* output write buffer overflow? */
                {
                    local_retval = NOERROR; /* local return value, (so ERDIN = "Error reading input file" will not go to spclose () on a SIGINT = ^C or SIGTERM "edge" in this routine,) assume no error */

                    switch (inchar = inbuffer[j]) /* get the input character */
                    {

                        case 0x04: /* end of transaction to be printed to query file, 0 = no, 1 = yes */

                            eot_flag = 1; /* set the end of transaction to be printed to query file, 0 = no, 1 = yes */

                        case '\0': /* null character for safety */
                        case '\n': /* EOL character? */
                        case ';':  /* end of statement character? */

                            if (ref > &outbuffer[0]) /* anything to interpret? */
                            {
                                *ref = '\0'; /* terminate the record */
                                local_retval = interpret (fdd, outbuffer, &quit_flag, query_name); /* interpret the command */
                            }

                            if (eot_flag == 1) /* end of transaction to be printed to query file, 0 = no, 1 = yes? */
                            {
                                local_retval = EQWRITE; /* assume error writing to query file */

                                if ((local_retval = write_query (&EOT, (size_t) 1)) == NOERROR) /* write the EOT character to the query file to mark end of transmission/transaction */
                                {
                                    local_retval = NOERROR; /* assume no error */
                                }

                                eot_flag = 0; /* reset end of transaction to be printed to query file, 0 = no, 1 = yes */
                            }

                            ref = &outbuffer[0]; /* first character in output write buffer */
                            break;

                        default:

                            *ref ++ = tolower (inchar); /* save the lower case of the character in the write output buffer, next character */
                            break;

                    }

                }

            }

        }

        if (cnt == 0) /* no characters? */
        {
            sleep (1); /* infiled is probably a pipe, or a network device descriptor, wait one second for latency */
        }

        else if (cnt < 0) /* read () error, (cnt == 0 is normal EOF, cnt == -1 a read error)? */
        {
            local_retval = ERDIN; /* assume error reading input file */
        }

    }

    return (retval = local_retval); /* return any errors */
}

/*

Interpret a command.

int interpret (int fdd, char *buffer, int *quit_flag, char *query_name);

I) Data structure:

    A) The serial device descriptor, fdd, should be open.

    B) The record, buffer, containing the single command should be an
       EOS = '\0' terminated string of characters-white space
       separates the tokens, (named "fields" in the comments.)

    C) The quit_flag is a reference to an int in parent_write () that
       will indicate the program is to be shutdown.

    D) The query file name, query_name, should be NULL if stderr is
       used for the query file, or reference an EOS terminated
       character string of the file name.

II) Execution:

    A) The interpret () function is recursive:

        1) If an "if" construct, (ISSET, ISNOTSET,) is encountered, a
           recursive call will be made to determine the action.

    B) There are two types of commands:

        1) Variable commands, (SET ABC, RESET ABC, ISSET ABC, etc.)

            a) Which deals with varibles.

        2) X10 Commands, (on, off, dim, etc.)

            a) Which are sent to the serial device descriptor, fdd.

    C) Both types of commands are installed in a single hash lookup
       table initialised in parent_write (). This hash table also
       contains the names of set variables, and time
       stamps. (Variables are either in the hash table, meaning set,
       or not, meaning reset.)

    D) The input buffer will be parsed into EOS = '\0' terminated
       character strings of keywords of both commands and variables,
       which will be looked up in the hash table.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

int interpret (int fdd, char *buffer, int *quit_flag, char *query_name)

#else

int interpret (fdd, buffer, query_name)
int fdd;
char *buffer;
int *quit_flag;
char *query_name;

#endif

{
    char parsebuffer[BUFLEN], /* parsed i/o buffer */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         token_separators[] = TOKEN_SEPARATORS, /* file record field separators */
         strbuffer[BUFLEN], /* intermediate string buffer */
         *timestamp, /* time stamp */
         results[BUFLEN]; /* results character string, for writing to the query file */

    int local_retval = NOERROR, /* return value, assume no error */
        fields, /* number of fields in a record */
        x10_command, /* x10 command flag, 0 = no, 1 = yes */
        iterations, /* number of DIM or BRIGHT commands to issue */
        intensity_cmd, /* command a DIM or BRIGHT command, 0 = no, 1 = yes */
        diff_time, /* the difference between the current time and time a variable was set */
        i; /* general counter */

    unsigned char housecode, /* the house code */
                  device; /* the unit code */

    time_t time_stamp; /* ctime for time stamps */

    HASH *element;  /* reference to hash table element */

    pid_t pid_exec; /* pid of system () call */

    if ((fields = strtoken (buffer, parsebuffer, token, token_separators)) > 0) /* parse the record into fields, skip blank records */
    {

        if (token[0][0] != '#') /* if the first character of the first field is a '#' character, skip it */
        {

            if ((element = hash_find (&hash_variable, token[0])) != (HASH *) 0) /* find the key in the hash table; if found, its a non-x10 command */
            {

                switch (element->command) /* which command */
                {

                    case QUIT:

                        *quit_flag = 1; /* yes, set the quit flag */
                        break;

                    case EXIT:

                        *quit_flag = 1; /* yes, set the quit flag */
                        break;

                    case SET:

                        if (fields > 1) /* at least two fields? */
                        {
                            (void) hash_delete (&hash_variable, token[1]); /* delete any previous identical key */
                            local_retval = hash_insert (&hash_variable, token[1]); /* insert the key into the hash table */
                        }

                        else
                        {
                            (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                        }

                        break;

                    case RESET:
                    case UNSET:

                        if (fields > 1) /* at least two fields? */
                        {
                            (void) hash_delete (&hash_variable, token[1]); /* delete any previous same key */
                        }

                        else
                        {
                            (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                        }

                        break;

                    case ISSET:

                        if (fields > 1) /* at least two fields? */
                        {

                            if ((element = hash_find (&hash_variable, token[1])) != (HASH *) 0) /* find a hash table element with a key, data, in the hash table */
                            {

                                if (fields > 2) /* found it, the variable is set, at least three fields? */
                                {
                                    local_retval = interpret (fdd, buffer + (token[2] - parsebuffer), quit_flag, query_name); /* interpret the remainder of the command */
                                }

                                else /* nothing to do, it was a request for status of the variable, print it */
                                {
                                    (void) strcpy (results, token[1]); /* save the variable name */
                                    (void) strcat (results, " is set\n"); /* save the variable's state */
                                    local_retval = write_query (results, strlen (results)); /* write the results to the query file */
                                }

                            }

                            else if (fields == 2) /* didn't find it, only field is a request to test whether set, or not, it isn't, print it */
                            {
                                (void) strcpy (results, token[1]); /* save the variable name */
                                (void) strcat (results, " is not set\n"); /* save the variable's state */
                                local_retval = write_query (results, strlen (results)); /* write the results to the query file */
                            }

                        }

                        else
                        {
                            (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                        }

                        break;

                    case ISNOTSET:

                        if (fields > 1) /* at least two fields? */
                        {

                            if ((element = hash_find (&hash_variable, token[1])) == (HASH *) 0) /* find a hash table element with a key, data, in the hash table */
                            {

                                if (fields > 2) /* didn't find it, the variable is not set, at least three fields? */
                                {
                                    local_retval = interpret (fdd, buffer + (token[2] - parsebuffer), quit_flag, query_name); /* interpret the remainder of the command */
                                }

                                else /* nothing to do, it was a request for status of the variable, print it */
                                {
                                    (void) strcpy (results, token[1]); /* save the variable name */
                                    (void) strcat (results, " is not set\n"); /* save the variable's state */
                                    local_retval = write_query (results, strlen (results)); /* write the results to the query file */
                                }

                            }

                            else if (fields == 2) /* found it, only field is a request to test whether set, or not, it is, print it */
                            {
                                (void) strcpy (results, token[1]); /* save the variable name */
                                (void) strcat (results, " is set\n"); /* save the variable's state */
                                local_retval = write_query (results, strlen (results)); /* write the results to the query file */
                            }

                        }

                        else
                        {
                            (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                        }

                        break;

                    case TIMESET:

                        if (fields > 1) /* at least two fields? */
                        {

                            if ((element = hash_find (&hash_variable, token[1])) != (HASH *) 0) /* find a hash table element with a key, data, in the hash table */
                            {
                                (void) sprintf (results, "%d\t", (int) element->time_set); /* format the ctime string */
                                (void) strcat (results, token[1]); /* save the variable name */
                                (void) strcat (results, "\n"); /* add an EOL */
                                local_retval = write_query (results, strlen (results)); /* write the results to the query file */
                            }

                        }

                        else
                        {
                            (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                        }

                        break;

                    case ISTIMESET:

                        if (fields > 4) /* at least four fields? */
                        {

                            if ((element = hash_find (&hash_variable, token[1])) != (HASH *) 0) /* find a hash table element with a key, data, in the hash table */
                            {

                                if ((diff_time = atoi (token[2])) != 0) /* get the time differential between the current time and the time the variable was set; note atoi() returns zero for non-valid, i.e., non-integer, arguments; an argument of an integer value means now, meaning always true */
                                {
                                    local_retval = ETIME; /* assume error getting time */

                                    if ((time_stamp = time ((time_t *) 0)) != -1) /* get the time */
                                    {
                                        local_retval = NOERROR; /* assume no error */

                                        if (diff_time > 0) /* request for enough elapsed time? */
                                        {

                                            if ((time_stamp - element->time_set) >= diff_time) /* time differential between the current time and the time the variable was set greater than zero? */
                                            {
                                                local_retval = interpret (fdd, buffer + (token[3] - parsebuffer), quit_flag, query_name); /* interpret the remainder of the command */
                                            }

                                        }

                                        else /* diff_time < 0, request for not enough elapsed time */
                                        {

                                            if ((time_stamp - element->time_set) <= -diff_time) /* time differential between the current time and the time the variable was set greater than zero? */
                                            {
                                                local_retval = interpret (fdd, buffer + (token[3] - parsebuffer), quit_flag, query_name); /* interpret the remainder of the command */
                                            }

                                        }

                                    }

                                }

                                else
                                {
                                    (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                                }

                            }

                            else /* nothing to do, the variable is not set, so doesn't have a set time */
                            {

                                if (verbose == 1) /* verbose diagnostics, 0 = no, 1 = yes? */
                                {
                                    (void) strcpy (results, token[1]); /* save the variable name */
                                    (void) strcat (results, " is not set\n"); /* save the variable's state */
                                    local_retval = write_query (results, strlen (results)); /* write the results to the query file */
                                }

                            }

                        }

                        else
                        {
                            (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                        }

                        break;

                    case EXEC:
                    case SYSTEM:

                        if (fields > 1) /* at least two fields? */
                        {

                            (void) fprintf (stderr, "Executing: %s\n", buffer + (token[1] - parsebuffer)); /* print the info */

                            if (signal (SIGCHLD, SIG_IGN) != SIG_ERR) /* ignore signals in child processes */
                            {

                                if ((pid_exec = fork ()) < 0) /* fork the system process */
                                {
                                    (void) fprintf (stderr, "%s", error_message[-EFORK]); /* error error forking to the system process, print the error and continue */
                                }

                                else if (pid_exec == 0) /* system process or parent_write process? */
                                {

                                    if (system (buffer + (token[1] - parsebuffer)) == -1) /* execute the system command */
                                    {
                                        (void) fprintf (stderr, "%s", error_message[-ESYS]); /* error error system call, print the error and continue */
                                    }

                                    exit (0); /* exit the system process */

                                }

                                else
                                {

                                    if (waitpid (pid_exec, 0, 0) == -1) /* wait for the system process command to terminate */
                                    {
                                        (void) fprintf (stderr, "%s", error_message[-ESYS]); /* error waiting for system call, print the error and continue */
                                    }

                                }

                                if (signal (SIGCHLD, spclose) == SIG_ERR) /* install the interrupt handler for SIGCHLD */
                                {
                                    (void) fprintf (stderr, "%s", error_message[-ESIG]); /* error installing signal handler, print the error and continue */
                                    (void) fprintf (stderr, "Attempting to exit ...\n"); /* and notifiy */
                                    (void) kill (pid_exec, SIGTERM); /* make sure the system process command is not running, kill the system process command */
                                    *quit_flag = 1; /* set the quit flag to shut down the parent_write and parent_read processes in an orderly manner as possible */
                                }

                            }

                            else
                            {
                                (void) fprintf (stderr, "%s", error_message[-ESIG]); /* error installing signal handler, print the error and continue */
                            }

                        }

                        else
                        {
                            (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                        }

                        break;

                    default: /* not a recognized command */

                        (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                        break;

                }

            }

            else if (fields > 1) /* possibly an x10 command which requires at least two fields? */
            {
                x10_command = 0; /* x10 command flag, 0 = no, 1 = yes, assume not */
                iterations = 1; /* number of DIM or BRIGHT commands to issue, assume 1 */
                intensity_cmd = 0, /* reset the command a DIM or BRIGHT command, 0 = no, 1 = yes */
                device = 0; /* the unit code, assume ALL_OFF, ALL_ON, ALL_LAMPS_OFF, ALL_LAMPS_ON, i.e., device 0 */

                if ((element = hash_find (&hash_variable, token[1])) != (HASH *) 0) /* find a hash table element with a key, data, in the hash table */
                {

                    switch (element->command) /* which command */
                    {

                        case DIM:
                        case BRIGHT:

                            if (fields > 2) /* at least three fields? */
                            {
                                iterations = atoi (token[2]); /* get the iterations */
                            }

                            if (token[0][1] == '\0') /* DIM or BRIGHT command contain house code only, (for compatability with X10 commands)? */
                            {
                                device = (unsigned char) 0; /* yes, set the device code to zero */
                            }

                            else
                            {
                                device = (unsigned char) atoi (&token[0][1]) - 1; /* get the device/unit code */
                            }

                            x10_command = 1; /* x10 command flag, 0 = no, 1 = yes, yes */
                            intensity_cmd = 1; /* set the command a DIM or BRIGHT command, 0 = no, 1 = yes */
                            break;

                        case ALL_OFF:
                        case ALL_ON:
                        case ALL_LAMPS_OFF:
                        case ALL_LAMPS_ON:

                            x10_command = 1; /* x10 command flag, 0 = no, 1 = yes, yes */
                            break;

                        case ON:
                        case OFF:

                            device = (unsigned char) atoi (&token[0][1]) - 1; /* get the device/unit code */
                            x10_command = 1; /* x10 command flag, 0 = no, 1 = yes, yes */
                            break;

                        default:

                            (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */

                    }

                    if (x10_command == 1) /* x10 command flag, 0 = no, 1 = yes, set? */
                    {
                        local_retval = EX10; /* assume error x10 command */

                        if ((housecode = (unsigned char) toupper (token[0][0]) - 'A') < 16) /* house codes must be 0 through 16 inclusive, (its unsigned, so 0 need not be tested for negative values) */
                        {

                            if (device < 16) /* unit codes must be 0 through 16 inclusive, (its unsigned, so 0 need not be tested for negative values) */
                            {

                                if ((iterations >= 0) && (iterations < 13)) /* iterations must be between 0 and 20, zero means do nothing */
                                {
                                    local_retval = NOERROR; /* assume no error */

                                    if (intensity_cmd == 1) /* command a DIM or BRIGHT command, 0 = no, 1 = yes? */
                                    {

                                        if (verbose == 1) /* verbose diagnostics, 0 = no, 1 = yes? */
                                        {
                                            local_retval = ETIME; /* assume error getting time */

                                            if ((time_stamp = time ((time_t *) 0)) != -1) /* get the time */
                                            {

                                                if ((timestamp = ctime (&time_stamp)) != (char *) 0) /* format the local time */
                                                {
                                                    (void) strcpy (strbuffer, buffer); /* save the X10 command */
                                                    (void) strcat (strbuffer, "\n"); /* and add an EOL */
                                                    local_retval = write_outfile (strbuffer, strlen (strbuffer), timestamp); /* write the record to the output file */
                                                }

                                            }

                                        }

                                        if (local_retval == NOERROR) /* any errors? */
                                        {

                                            if ((device == (unsigned char) 0) || ((local_retval = cm17a_out (fdd, device, housecode, ON)) == NOERROR)) /* issue the ON command to set the address, but only if there was a device code, (for compatability with X10 commands) */
                                            {

                                                for (i = 0; (i < iterations) && (local_retval == NOERROR); i ++) /* for each iteration */
                                                {
                                                    local_retval = ETIME; /* assume error getting time */

                                                    if ((time_stamp = time ((time_t *) 0)) != -1) /* get the time */
                                                    {

                                                        if ((timestamp = ctime (&time_stamp)) != (char *) 0) /* format the local time */
                                                        {
                                                            (void) strcpy (strbuffer, buffer); /* save the X10 command */
                                                            (void) strcat (strbuffer, "\n"); /* and add an EOL */
                                                            local_retval = write_outfile (strbuffer, strlen (strbuffer), timestamp); /* write the record to the output file */
                                                        }

                                                    }

                                                    if (local_retval == NOERROR) /* any errors? */
                                                    {
                                                        local_retval = cm17a_out (fdd, device, housecode, element->command); /* issue a single DIM/BRIGHT command */
                                                    }

                                                }

                                            }

                                        }

                                    }

                                    else
                                    {

                                        if (verbose == 1) /* verbose diagnostics, 0 = no, 1 = yes? */
                                        {
                                            local_retval = ETIME; /* assume error getting time */

                                            if ((time_stamp = time ((time_t *) 0)) != -1) /* get the time */
                                            {

                                                if ((timestamp = ctime (&time_stamp)) != (char *) 0) /* format the local time */
                                                {
                                                    (void) strcpy (strbuffer, buffer); /* save the X10 command */
                                                    (void) strcat (strbuffer, "\n"); /* and add an EOL */
                                                    local_retval = write_outfile (strbuffer, strlen (strbuffer), timestamp); /* write the record to the output file */
                                                }

                                            }

                                        }

                                        if (local_retval == NOERROR) /* any errors? */
                                        {
                                            local_retval = cm17a_out (fdd, device, housecode, element->command); /* issue the command */
                                        }

                                    }

                                }

                            }

                        }

                        if (local_retval == EX10) /* any errors? */
                        {
                            (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
                            local_retval = NOERROR; /* assume no error */
                        }

                    }

                }

            }

            else /* none of the above */
            {
                (void) fprintf (stderr, "Error: %s\n", buffer); /* no such command, print the error and continue */
            }

        }

    }

    return (retval = local_retval); /* return any errors */
}

/*

Write a record to the query file.

int write_query (char *buffer, size_t len);

I) Data structure:

    A) The buffer should contain an EOS terminated character string of
       length, len.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int write_query (char *buffer, size_t len)

#else

static int write_query (buffer, len)
char *buffer;
size_t len;

#endif

{
    int local_retval = EQWRITE; /* assume error writing to query file */

    if (write (queryfile, buffer, len) == (ssize_t) len) /* write the record to the query file */
    {
        local_retval = NOERROR; /* assume no error */
    }

    return (retval = local_retval); /* return any errors */
}

/*

Write a record to the output file.

int write_outfile (char *outbuf, size_t buflen, char *timestamp);

I) Data structure:

    A) The buffer should contain an EOS terminated string of chracters
       of length buflen.

    B) The time string, timestamp, should be an EOS terminated string
       of characters, terminated with an EOS, and produced by
       ctime(3).

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int write_outfile (char *outbuf, size_t buflen, char *timestamp)

#else

static int write_outfile (outbuf, buflen, timestamp)
char *outbuff;
size_t buflen;
char *timestamp;

#endif

{
    int local_retval = EOWRITE; /* local return value, assume error writing to output file */

    size_t time_len; /* length of time stamp */ /* character counter */

    time_len = strlen (&timestamp[4]); /* get the length of the time stamp */
    time_len --; /* remove the '\n' character from the time stamp */
    local_retval = EOWRITE; /* assume error writing to output file */

    if (write (outfile, &timestamp[4], time_len) > 0) /* write the time stamp to the output device */
    {

        if (write (outfile, "\t", 1) > 0) /* add a tab time stamp delimiter */
        {

            if (write (outfile, outbuf, buflen) > 0) /* write the formatted output to the output device */
            {

                if (outfile == STDOUT_FILENO) /* output file stdout, (can't flush stdout)? */
                {
                    local_retval = NOERROR; /* assume no error */
                }

                else
                {
                    local_retval = EOFLUSH; /* assume error flushing output file */

                    if (fsync (outfile) == 0) /* flush the output file */
                    {
                        local_retval = NOERROR; /* assume no error */
                    }

                }

            }

        }

    }

    return (retval = local_retval); /* return any errors */
}
