/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2008, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

spx10delay.c is for providing interoperability between the input file
of the spx10S0 program and externally applied commands to be executed
after a time delay-the spx10S0 program writes internally generated
commands to its input file, which would create concurrency issues with
externally generated commands; the spx10delay, spx10lock and spx10S0
programs use fcntl(2) advisory locking of the spx10S0 input file for
interoperability between internally generated commands within the
spx10S0 program, and externally generated commands by the spx10lock
and spx10delay programs.

Only text transactions are permitted, (the program uses the EOT = ^D =
end of transmission/transaction character for EOF and transaction
concurrency control across the link, which is not compatible with
binary data, as in telnet(1).) An EOT = ^D = end of
transmission/transaction character character is appended to each
record in the output file.

An appropriate example usage would be:

    mknod infile p

    spx10S0 -f spx10S0db -i infile -q queryfile
    spx10delay -o infile -i commands

The spx10delay program writes externally supplied, (via stdin,) x10
commands to the named pipe, infile, and the spx10S0 program reads the
commands from the pipe.

The structure of the commands is one command per EOL terminated
record, which starts with the delay before the command is to be
executed in seconds, exactly one tab character, and the command to be
executed.

$Revision: 0.0 $
$Date: 2008/01/18 19:36:00 $
$Id: spx10delay.c,v 0.0 2008/01/18 19:36:00 john Exp $
$Log: spx10delay.c,v $
Revision 0.0  2008/01/18 19:36:00  john
Initial version


*/

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#define BUFLEN BUFSIZ /* size of record buffers */

#define TOKEN_SEPARATORS "\t\n\r\b," /* input file record field separators */

#define MAX_DELAYS 256 /* maximum number of delayed commands pending at any instant */

static char rcsid[] = "$Id: spx10delay.c,v 0.0 01/18/08 21:31:06 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2008, John Conover, All Rights Reserved"; /* the copyright banner */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Read from a file, write to a file\n",
    "Usage: spx10delay [-i infilename] [-o outfilename] [-V] [-v]\n",
    "    -i infilename, input filename, (stdin)\n",
    "    -o outfilename, output file name, (stdout)\n",
    "    -V, verbose output\n",
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error opening output file\n",
    "Error closing output file\n",
    "Error locking output file\n",
    "Error unlocking output file\n",
    "Error reading input file\n",
    "Error writing to output file\n",
    "Error syncing output file\n",
    "Error overflow buffer\n",
    "Error getting status of input file\n",
    "Error getting status of output file\n",
    "Error number of threads\n",
    "Error creating delay thread\n",
    "Error in command construction\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EIOPEN 2
#define EICLOSE 3
#define EOOPEN 4
#define EOCLOSE 5
#define EOLCK 6
#define EOULCK 7
#define EIREAD 8
#define EOWRITE 9
#define EOSYNC 10
#define EOVRFL 11
#define EISTAT 12
#define EOSTAT 13
#define ENUMTH 14
#define ETHREAD 15
#define ECMD 16

enum file_type /* input and output file type */
{
    REG, /* regular file */
    PIPE, /* pipe */
    TTY /* terminal, default */
};

typedef struct delay /* delay structure for passing data to threaded write_out () function */
{
    char *out_name; /* output file name */
    enum file_type out_type; /* output file type */
    char inbuffer[BUFLEN + 1]; /* copy of input read buffer */
    size_t cnt_out; /* count of characters to write */
    int *ref_thread_retval; /* reference to return value from a write_out () thread; if != NOERROR, a exception occurred in write_out (), and the program should exit at the next character read from the input file */
    pthread_t thread; /* thread identifier for the delay */
    pthread_attr_t tattr; /* thread attribute for the delay */
    int delay; /* number of seconds to delay before command is issued */
    int in_use; /* this array element in use, 0 = no, 1 = yes */
} DELAY;

static DELAY delay_array[MAX_DELAYS]; /* array of DELAY structures, one per delayed command */

static int verbose_flag = 0; /* verbose output flag, 0 = no, 1 = yes */

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static int strtoken (char *string, char *parse_array, char **parse, char *delim); /* parse a record based on sequential delimiters */
static void *write_out (void *arg); /* write the command to the output file */

#else

static void print_message (); /* print any error messages */
static int strtoken (); /* parse a record based on sequential delimiters */
static void *write_out (); /* write the command to the output file */

#endif

static char EOT = (char) 0x04; /* EOT = ^D, end of transmission/transaction marker character */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char *in_name = (char *) 0, /* input file name */
         *out_name = (char *) 0, /* output file name */
         inchar, /* character from input file */
         inbuffer[BUFLEN + 1], /* input read buffer */
         *ref, /* character reference in input buffer */
         *sentinel = &inbuffer[BUFLEN], /* reference to sentinel character in input buffer */
         token_separators[] = TOKEN_SEPARATORS, /* input file record field separators */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         *token[BUFLEN / 2]; /* reference to tokens in parsed i/o buffer */

    int retval = NOERROR, /* return value, assume no error */
        thread_retval = NOERROR, /* return value from a write_out () thread; if != NOERROR, a exception occurred in write_out (), and the program should exit at the next character read from the input file */
        infile = STDIN_FILENO, /* input file descriptor */
        input_mode = O_RDONLY, /* open mode for input file */
        fields, /* number of fields in an input file record */
        delay, /* number of seconds to delay before command is issued */
        i, /* general counter */
        c; /* command line switch */

    ssize_t cnt_in = 0, /* count of characters read */
            cnt_out; /* count of characters to write */

    struct stat file_stat; /* lstat () structure */

    enum file_type in_type = TTY, /* input file type */
                   out_type = TTY; /* output file type */

    *sentinel = '\0'; /* sentinel character in input buffer */

    while ((retval == NOERROR) && ((c = getopt (argc, argv, "i:o:vV")) != EOF)) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'i': /* request for input file name? */

                in_name = optarg; /* yes, save the input file name */
                retval = EISTAT; /* assume error getting status of input file */

                if (lstat (in_name, &file_stat) == 0) /* get the input file status */
                {

                    if (S_ISFIFO (file_stat.st_mode)) /* input file a pipe? */
                    {
                        in_type = PIPE;
                        input_mode = O_RDWR; /* open mode for input file */
                    }

                    else if (S_ISREG (file_stat.st_mode)) /* input file a regular file? */
                    {
                        in_type = REG;
                    }

                    /* else the input file type is defaulted to TTY */

                    retval = NOERROR; /* assume no error */
                }

                break;

            case 'o': /* request for output file name? */

                out_name = optarg; /* yes, save the output file name */
                retval = EOSTAT; /* assume error getting status of output file */

                if (lstat (out_name, &file_stat) == 0) /* get the output file status */
                {

                    if (S_ISFIFO (file_stat.st_mode)) /* output file a pipe? */
                    {
                        out_type = PIPE;
                    }

                    else if (S_ISREG (file_stat.st_mode)) /* output file a regular file? */
                    {
                        out_type = REG;
                    }

                    /* else the output file type is defaulted to TTY */

                    retval = NOERROR; /* assume no error */
                }

                else if (errno == ENOENT) /* a non-existent output file will be created by open () as a regular file, below; any other lstat () error is a hard error */
                {
                    out_type = REG;
                    retval = NOERROR; /* assume no error */
                }

                break;

            case 'V': /* request for verbose output, 0 = no, 1 = yes */

                verbose_flag = 1; /* yes, set the verbose output flag, 0 = no, 1 = yes */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    for (i = 0; i < MAX_DELAYS; i ++) /* for each element in the array of DELAY structures */
    {
        delay_array[i].out_name = out_name; /* save the output file name */
        delay_array[i].out_type = TTY; /* save the output file type */
        delay_array[i].inbuffer[0] = '\0'; /* null the copy of input read buffer */
        delay_array[i].inbuffer[BUFLEN] = '\0'; /* null the sentinel character in input buffer */
        delay_array[i].cnt_out = 0; /* save the count of characters to write */
        delay_array[i].ref_thread_retval = &retval; /* set the reference to return value from a write_out () thread; if != NOERROR, a exception occurred in write_out (), and the program should exit at the next character read from the input file */
        delay_array[i].thread = (pthread_t) 0; /* null the thread identifier for the delay */
        delay_array[i].delay = 0; /* null the number of seconds to delay before command is issued */
        delay_array[i].in_use = 0; /* reset the this array element in use, 0 = no, 1 = yes */
    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EIOPEN; /* assume error opening input file */

        if ((infile = (in_name == (char *) 0) ? infile : open (in_name, input_mode)) >= 0) /* open the input file */
        {
            ref = inbuffer; /* reset the character reference in input buffer */
            *ref = '\0'; /* terminate the record */
            retval = EIREAD; /* assume error reading input file */

            /* retval == EIREAD means no error in progress; any other error means to drop through with the error, (without reading the next record); an error in reading from the input file means drop through with the read error */

            while ((retval == EIREAD) && ((cnt_in = read (infile, &inchar, sizeof (char)) == sizeof (char)))) /* read characters from the input file */
            {

                if (thread_retval != NOERROR) /* return value from a write_out () thread; if != NOERROR, a exception occurred in write_out (), and the program should exit at the next character read from the input file */
                {
                    retval = thread_retval; /* save the return value from a write_out () thread */
                    break; /* shutdown */
                }

                if (inchar != EOT) /* skip EOT = ^D = end of transmission/transaction character from the input; an EOT will be appended to every record output */
                {
                    retval = EOVRFL; /* assume error overflow buffer */

                    if (ref < sentinel) /* buffer overflow? */
                    {
                        *ref = inchar; /* save the character in the input buffer */
                        ref ++; /* prepare for next character reference in input buffer */
                        *ref = '\0'; /* terminate the record; may overwrite the sentinel '\0' */
                        retval = EIREAD; /* assume error reading input file */

                        if (inchar == '\n') /* character an EOL? */
                        {
                            cnt_out = ref - inbuffer; /* save the count of characters read */
                            retval = ECMD; /* assume error in command construction */

                            if ((fields = strtoken (inbuffer, parsebuffer, token, token_separators)) == 2) /* parse the input buffer; there should be exactly two fields */
                            {

                                if ((delay = atoi (token[0])) > 0) /* save the number of seconds to delay before command is issued, which is the first field, with a sanity check */
                                {
                                    cnt_out = strlen (token[1]); /* save the count of characters in the second field, which is the command */
                                    retval = ENUMTH; /* assume error allocating memory */

                                    for (i = 0; i < MAX_DELAYS; i ++) /* for each element in the array of DELAY structures, find the first array element that is not in use */
                                    {

                                        if (delay_array[i].in_use == 0) /* array element in use, 0 = no, 1 = yes? */
                                        {
                                            retval = NOERROR; /* assume no error */
                                            break; /* found the first array element that is not in use */
                                        }

                                    }

                                    if (retval == NOERROR) /* any errors? */
                                    {
                                        (void) strcpy (delay_array[i].inbuffer, token[1]); /* save the second field, which is the command */
                                        delay_array[i].cnt_out = cnt_out; /* save the count of characters to write */
                                        delay_array[i].delay = delay; /* save the number of seconds to delay before command is issued */
                                        delay_array[i].in_use = 1; /* set the this array element in use, 0 = no, 1 = yes */
                                        retval = ETHREAD; /* assume error creating delay thread */

                                        if (verbose_flag == 1) /* verbose output flag, 0 = no, 1 = yes? */
                                        {
                                            (void) fprintf (stderr, "Delaying %d seconds: %s\n", delay, token[1]); /* print the record */
                                        }

                                        if (pthread_attr_init (&(delay_array[i].tattr)) == 0) /* initialize the thread with default attributes for this delay */
                                        {

                                            if (pthread_attr_setdetachstate (&(delay_array[i].tattr), PTHREAD_CREATE_DETACHED) == 0) /* set the thread to detach for this delay; the thread will wait for the specified delay, execute, and exit without a return code to this routine */
                                            {

                                                if (pthread_create (&(delay_array[i].thread), &(delay_array[i].tattr), write_out, (void *) &delay_array[i]) == 0) /* create the thread for this delay */
                                                {
                                                    retval = EIREAD; /* assume error reading input file */
                                                }

                                            }

                                        }

                                    }

                                }

                            }

                            ref = inbuffer; /* reset the character reference in input buffer */
                            *ref = '\0'; /* terminate the record */
                        }

                    }

                }

            }

            retval = ((cnt_in == 0) && (retval == EIREAD)) ? NOERROR : retval; /* a count of zero, and error reading input file was an EOF or ^D, a normal exit */

            if ((infile != STDIN_FILENO) && (close (infile) != 0)) /* close the input file, but only if its not stdin */
            {
                retval = (retval == NOERROR) ? EICLOSE : retval; /* assume error closing input file */
            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with any error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Parse a record based on sequential delimiters.

static int strtoken (char *string, char *parse_array, char **parse, char *delim);

Parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field.

Returns the number of fields parsed.

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string,
*parse_array,
**parse,
*delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}

/*
*/

#ifdef __STDC__

static void *write_out (void *arg)

#else

static void *write_out (arg)
void *arg;

#endif

{
    char *out_name, /* output file name */
         *inbuffer; /* input read buffer */

    int retval = EOOPEN, /* return value, assume error opening output file */
        outfile = STDOUT_FILENO, /* output file descriptor */
        *ref_thread_retval, /* return value from a write_out () thread; if != NOERROR, a exception occurred in write_out (), and the program should exit at the next character read from the input file */
        delay; /* number of seconds to delay before command is issued */

    size_t cnt_out; /* count of characters to write */

    struct flock lockfile; /* lock file structure */

    enum file_type out_type; /* output file type */

    DELAY *delay_data = (DELAY *) arg; /* reference to delay structure */

    out_name = delay_data->out_name; /* output file name */
    out_type = delay_data->out_type; /* output file type */
    inbuffer = delay_data->inbuffer; /* input read buffer */
    cnt_out = delay_data->cnt_out; /* count of characters to write */
    ref_thread_retval = delay_data->ref_thread_retval; /* return value from a write_out () thread; if != NOERROR, a exception occurred in write_out (), and the program should exit at the next character read from the input file */
    delay = delay_data->delay; /* number of seconds to delay before command is issued */

    (void) sleep (delay); /* wait for the specified delay */

    if ((outfile = (out_name == (char *) 0) ? outfile : open (out_name, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the output file */
    {
        lockfile.l_type = F_WRLCK; /* exclusive write lock the entire output file */
        lockfile.l_whence = (short int) SEEK_SET;
        lockfile.l_start = (off_t) 0;
        lockfile.l_len = (off_t) 0;
        retval = EOLCK; /* assume error opening output file */

        if ((out_type == TTY) || (fcntl (outfile, F_SETLKW, &lockfile) == 0)) /* exclusive write lock the output file, but only if its a regular file or pipe */
        {

            if (verbose_flag == 1) /* verbose output flag, 0 = no, 1 = yes? */
            {
                (void) fprintf (stderr, "Executing: %s\n", inbuffer); /* print the record */
            }

            retval = EOWRITE; /* assume error writing to output file */

            if ((write (outfile, inbuffer, cnt_out)) == (ssize_t) cnt_out) /* write the characters to the output device */
            {

                if ((write (outfile, "\n", 1)) == (ssize_t) 1) /* write an EOL record terminator to the output device */
                {

                    if ((out_type != PIPE) || (write (outfile, &EOT, (size_t) 1) == (ssize_t) 1)) /* write the EOT = ^D, end of transmission/transaction marker character to the output file */
                    {
                        retval = NOERROR; /* assume no error */

                        if ((out_type == REG) && (fsync (outfile) != 0)) /* flush the output file's buffers, but only if its a regular file; do not flush pipes or ttys */
                        {
                            retval = EOSYNC; /* assume error syncing output file */
                        }

                    }

                }

            }

            lockfile.l_type = F_UNLCK; /* unlock the file */

            if ((out_type != TTY) && (fcntl (outfile, F_SETLKW, &lockfile)) != 0) /* unlock the output file, but only if its a regular file or pipe */
            {
                retval = (retval == NOERROR) ? EOULCK : retval; /* assume error unlocking output file */
            }

        }

        if ((outfile != STDOUT_FILENO) && (close (outfile) != 0)) /* close the output file, but only if it is not stdout */
        {
            retval = (retval == NOERROR) ? EOCLOSE : retval; /* assume error closing output file */
        }

    }

    if (retval != NOERROR) /* any errors? */
    {
        *ref_thread_retval = retval; /* return value from a write_out () thread; if != NOERROR, a exception occurred in write_out (), and the program should exit at the next character read from the input file */
    }

    delay_data->in_use = 0; /* reset the this array element in use, 0 = no, 1 = yes */
    pthread_exit ((void *) 0); /* return nothing */
}
