/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

spx10lock.c is for providing interoperability between the input file
of the spx10S0 program and externally applied commands-the spx10S0
program writes internally generated commands to its input file, which
would create concurrency issues with externally generated commands;
both the spx10lock and spx10S0 programs use fcntl(2) advisory locking
of the spx10S0 input file for interoperability between internally
generated commands within the spx10S0 program, and externally
generated commands by the spx10lock program.

Queries are submitted to the spx10lock program on stdin.  The results
to queries are printed to stdout.

Multiple '\n' terminated queries can be submitted in a single
transaction to the input file of the spx10lock program, and all
queries in the transaction processed to stdout before the input file
lock is released.

Only text transactions are permitted, (the program uses the EOT = ^D =
end of transmission/transaction character for EOF and transaction
concurrency control across the link, which is not compatible with
binary data, as in telnet(1).)

If the program is invoked with arguments, they are sent to outfile,
instead of the input file being read.

An appropriate example usage would be:

    mknod infile p

    spx10S0 -f spx10S0db -i infile -q queryfile
    spx10lock -o infile -q queryfile -i commands

or

    spx10lock -o infile -q queryfile "command1" "command2" ...

The spx10lock program writes externally supplied, (via stdin,) x10
commands to the named pipe, infile, and the spx10S0 program reads the
commands from the pipe, and writes resulting data to queryfile, which
is read by the spx10lock program and printed to stdout.

To provide TCP/Ethernet connectivity, the ucspi tcpserver/tcpclient
programs at http://cr.yp.to/ucspi-tcp.html can be used for transaction
processing between two spx10lock programs; for example, on the server
machine, machine1:

    mknod infile p

    spx10S0 -f spx10S0db -i infile -q queryfile
    tcpserver -v 0 23456 spx10lock -r -o infile -q queryfile

and, on the client machine:

    tcpclient -v machine1 23456 spx10lock -u "a5 on"

would turn device a5 on, using TCP socket 23456.

The ucspi-tcp programs are robust network programs with enhanced
security features.

$Revision: 0.0 $
$Date: 2007/01/18 19:36:00 $
$Id: spx10lock.c,v 0.0 2007/01/18 19:36:00 john Exp $
$Log: spx10lock.c,v $
Revision 0.0  2007/01/18 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

static char rcsid[] = "$Id: spx10lock.c,v 0.0 2007/01/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Query spx10S0 variables/commands\n",
    "Usage: spx10lock [-i infilename] [-l] [-o oufilename]\n",
    "                 [-q queryfilename] [-r] [-u] [-v] [commands ...]\n",
    "    -i infilename, input filename, (stdin)\n",
    "    -l, limit input transaction size to BUFSIZE characters\n",
    "    -o outfilename, output filename, (stdout)\n",
    "    -q queryfilename, query file name, (stderr)\n",
    "    -r, do not remove EOT transmission/transaction marker\n",
    "    -u, ucspi-tcp client file descriptors\n",
    "    -v, print the program's version information\n",
    "    commands ..., list of commands instead of input file\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error reading input file\n",
    "Error opening output file\n",
    "Error closing output file\n",
    "Error writing to output file\n",
    "Error locking output file\n",
    "Error unlocking output file\n",
    "Error opening query file\n",
    "Error closing query file\n",
    "Error positioning query file\n",
    "Error reading query file\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPENIN 2
#define ECLOSEIN 3
#define ERDIN 4
#define EOPENOUT 5
#define ECLOSEOUT 6
#define EWRITEOUT 7
#define EOLCK 8
#define EOULCK 9
#define EQOPEN 10
#define EQCLOSE 11
#define EQPOS 12
#define EQRD 13

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */

#else

static void print_message (); /* print any error messages */

#endif

static char EOT = (char) 0x04; /* EOT = ^D, end of transmission/transaction marker character */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN + 1], /* i/o buffer */
         *sentinel = &buffer[BUFLEN], /* reference to sentinel character in i/o buffer */
         *infile_name = (char *) 0, /* reference to input file name */
         *outfile_name = (char *) 0, /* reference to output file name */
         *query_name = (char *) 0, /* reference to query file name */
         *eof; /* reference to EOF in buffer */

    int retval = NOERROR, /* return value, assume no error */
        c, /* command line switch */
        r_tran = 0, /* do not remove EOT transmission/transaction marker, 0 remove it, 1 = leave it */
        outfile = STDOUT_FILENO, /* output file descriptor */
        queryfile = STDERR_FILENO, /* query file descriptor */
        infile = STDIN_FILENO, /* input file descriptor */
        quit_flag = 0, /* flag to stop reading read_in after current write to write_out, 0 = no, 1 = yes */
        limit_trans = 0, /* limit the transaction size to one buffer read, 0 = no, 1 = yes */
        i; /* command line argument counter */

    size_t len = 0; /* record length */

    ssize_t cnt = 0, /* count of characters read */
            dummy; /* return value from write(2) to stdout, lint issue */

    struct flock locallock; /* output file locking structure */

    locallock.l_type = F_WRLCK; /* lock/unlock EOF of file for write */
    locallock.l_whence = (short int) SEEK_SET;
    locallock.l_start = (off_t) 0;
    locallock.l_len = (off_t) 0;

    *sentinel = '\0'; /* i/o buffer sentinel */

    while ((c = getopt (argc, argv, "i:lo:q:ruv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'i': /* request for input file name? */

                infile_name = optarg; /* yes, save the input file name */
                break;

            case 'l': /* request to limit the transaction size to one buffer read, 0 = no, 1 = yes? */

                limit_trans = 1; /* yes, set the limit the transaction size to one buffer read, 0 = no, 1 = yes */
                break;

            case 'o': /* request for output file name? */

                outfile_name = optarg; /* yes, save the output file name */
                break;

            case 'q': /* request for query file name? */

                query_name = optarg; /* yes, save the query file name */
                break;

            case 'r': /* request for do not remove EOT transmission/transaction marker, 0 remove it, 1 = leave it */

                r_tran = 1; /* yes, set the do not remove EOT transmission/transaction marker, 0 remove it, 1 = leave it */
                break;

            case 'u': /* request for ucspi file descriptors */

                outfile = 6; /* output file descriptor */
                queryfile = 7; /* query file descriptor */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EOPENOUT; /* assume error opening output file */

        if ((outfile = (outfile_name == (char *) 0) ? outfile : open (outfile_name, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the output file */
        {
            locallock.l_type = F_WRLCK; /* lock EOF of file for write */
            retval = EOLCK; /* assume error locking output file */

            if (fcntl (outfile, F_SETLKW, &locallock) == 0) /* write lock the EOF of the output file */
            {
                retval = EQOPEN; /* assume error opening query file */

                if ((queryfile = (query_name == (char *) 0) ? queryfile : open (query_name, O_RDONLY)) >= 0) /* open the query file */
                {
                    retval = EQPOS; /* assume error positioning to EOF of query file */

                    if ((queryfile == STDERR_FILENO) || (queryfile == 7) || (queryfile == 7) || (lseek (queryfile, (off_t) 0, SEEK_END) != (off_t) -1)) /* position to EOF of the query file */
                    {
                        retval = NOERROR; /* assume no error */

                        if (argc - optind > 0) /* queries as command line arguments? */
                        {

                            for (i = optind; (retval == NOERROR) && (i < argc); i ++) /* for each command line argument */
                            {
                                (void) strcpy (buffer, argv[i]); /* save a copy of the command line argument */
                                (void) strcat (buffer, "\n"); /* terminate the command line argument with an carriage return */
                                len = strlen (buffer); /* save the record length */
                                retval = EWRITEOUT; /* assume error writing to output file */

                                if (write (outfile, buffer, len) == (ssize_t) len) /* write the record to the output file */
                                {
                                    retval = NOERROR; /* assume no error */
                                }

                            }

                            if (write (outfile, &EOT, (size_t) 1) != (ssize_t) 1) /* write the EOT = ^D, end of transmission/transaction marker character to the output file */
                            {
                                retval = (retval == NOERROR) ? EWRITEOUT : retval; /* assume error writing to output file */
                            }

                        }

                        else
                        {
                            retval = EOPENIN; /* assume error opening input file */

                            if ((infile = (infile_name == (char *) 0) ? infile : open (infile_name, O_RDONLY)) >= 0) /* open the input file */
                            {
                                retval = NOERROR; /* assume no error */

                                while ((retval == NOERROR) && (quit_flag == 0) && ((cnt = read (infile, buffer, BUFLEN)) > 0)) /* read characters from the input file */
                                {
                                    buffer[cnt] = '\0'; /* terminate the characters with an EOS, (if the buffer is full, this will overwrite the existing EOS sentinel) */

                                    if ((eof = strchr (buffer, EOT)) != (char *) 0) /* buffer contain an EOT character? */
                                    {
                                        *eof = '\0'; /* remove the EOT character */
                                        cnt = eof - buffer; /* adjust the number of bytes read from input file to exclude the EOT marker */
                                        quit_flag = 1; /* schedule a stop after current buffer is written to write_out */
                                    }

                                    retval = EWRITEOUT; /* assume error writing to output file */

                                    if (write (outfile, buffer, cnt) == (ssize_t) cnt) /* write the record to the output file */
                                    {
                                        retval = NOERROR; /* assume no error */
                                    }

                                    if (limit_trans == 1) /* set the limit the transaction size to one buffer read, 0 = no, 1 = yes? */
                                    {
                                        break; /* no further reads from input file */
                                    }

                                }

                                if (cnt < 0) /* read () error, (cnt == 0 is normal EOF, cnt == -1, a read error)? */
                                {
                                    retval = (retval == NOERROR) ? ERDIN : retval; /* assume error reading input file */
                                }

                                if (write (outfile, &EOT, (size_t) 1) != (ssize_t) 1) /* write the EOT = ^D, end of transmission/transaction marker character to the output file */
                                {
                                    retval = (retval == NOERROR) ? EWRITEOUT : retval; /* assume error writing to output file */
                                }

                                if (infile != STDIN_FILENO) /* input file open? */
                                {

                                    if (close (infile) != 0) /* close the input file */
                                    {
                                        retval = (retval == NOERROR) ? ECLOSEIN : retval; /* assume error closing input file */
                                    }

                                    infile = STDIN_FILENO; /* input file descriptor */
                                }

                            }

                        }

                        if (retval == NOERROR) /* any errors? */
                        {
                            quit_flag = 0; /* flag to stop reading read_in after current write to write_out, 0 = no, 1 = yes */

                            while ((retval == NOERROR) && (quit_flag == 0) && ((cnt = read (queryfile, buffer, BUFLEN)) >= 0)) /* read characters from the input file, (until the EOT character is found) */
                            {

                                if (cnt == 0) /* any characters? */
                                {
                                    sleep (1); /* no, wait one second for the data to be appended to the query file which might have an EOT character */
                                }

                                else
                                {
                                    buffer[cnt] = '\0'; /* terminate the characters with an EOS, (if the buffer is full, this will overwrite the existing EOS sentinel) */

                                    if ((eof = strchr (buffer, EOT)) != (char *) 0) /* buffer contain an EOT character? */
                                    {

                                        if (r_tran == 0) /* do not remove EOT transmission/transaction marker, 0 remove it, 1 = leave it? */
                                        {
                                            *eof = '\0'; /* remove the EOT character */
                                            cnt = eof - buffer; /* adjust the number of bytes read from input file to exclude the EOT marker */
                                        }

                                        quit_flag = 1; /* schedule a stop after current buffer is written to stdout */
                                    }

                                    dummy = write (STDOUT_FILENO, buffer, cnt); /* write the characters to stdout */
                                }

                            }

                            if (cnt < 0) /* assume error reading query file */
                            {
                                retval = (retval == NOERROR) ? EQRD : retval; /* assume error reading input file */
                            }

                        }

                    }

                    if ((queryfile != STDERR_FILENO) && (queryfile != 7)) /* query file open? */
                    {

                        if (close (queryfile) != 0) /* close the query file */
                        {
                            retval = (retval == NOERROR) ? EQCLOSE : retval; /* assume error closing query file */
                        }

                        queryfile = STDERR_FILENO; /* query file descriptor */
                    }

                }

                locallock.l_type = F_UNLCK; /* unlock the file */

                if (fcntl (outfile, F_SETLKW, &locallock) != 0) /* unlock the input file of the parent_write() function */
                {
                    retval = (retval == NOERROR) ? EOULCK : retval; /* assume error unlocking file */
                }

            }

            if ((outfile != STDOUT_FILENO) && (outfile != 6)) /* using stdout as output? */
            {

                if (close (outfile) != 0) /* no, close the output file */
                {
                    retval = (retval == NOERROR) ? ECLOSEOUT : retval; /* assume error closing file */
                }

                outfile = STDOUT_FILENO; /* output file descriptor */
            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}
